#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
ParlAI Server file.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import importlib
import inspect
import json
import logging
import os
import time
import threading
import traceback
import asyncio
import sh
import shlex
import shutil
import subprocess
import uuid

import tornado.ioloop
import tornado.web
import tornado.websocket
import tornado.escape

from parlai.mturk.core.mturk_data_handler import MTurkDataHandler
from parlai.mturk.core.mturk_manager import MTurkManager
from parlai.mturk.webapp.run_mocks.mock_turk_manager import MockTurkManager
from typing import Dict, Any

# https://github.com/python/mypy/issues/1422
from parlai import __path__ as parlai_path  # type: ignore

parlai_path = parlai_path[0]

try:
    # https://github.com/python/mypy/issues/1422
    from parlai_internal import __path__ as parlai_int_path  # type: ignore

    parlai_int_path = parlai_int_path[0]
except Exception:
    parlai_int_path = None


DEFAULT_PORT = 8095
DEFAULT_HOSTNAME = "localhost"
DEFAULT_DB_FILE = 'pmt_data.db'
DEFAULT_SB_DB_FILE = 'pmt_sbdata.db'
IS_DEBUG = True

here = os.path.abspath(os.path.dirname(__file__))

tasks: Dict[Any, Any] = {}


def row_to_dict(row):
    return dict(zip(row.keys(), row))


def get_rand_id():
    return str(hex(int(time.time() * 10000000))[2:])


def force_dir(path):
    """
    Make sure the parent dir exists for path so we can write a file.
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)


def get_path(filename):
    """
    Get the path to an asset.
    """
    cwd = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    return os.path.join(cwd, filename)


def get_mturk_task_module(task_name, target_module):
    full_location = tasks[task_name]
    guess_loc = full_location.split('tasks/')[1]
    guess_class = '.'.join(guess_loc.split('/'))
    task_module = target_module.format(guess_class)
    base_format = 'parlai.mturk.tasks.{}'
    if 'parlai_internal' in full_location:
        base_format = 'parlai_internal.mturk.tasks.{}'
    find_location = base_format.format(task_module)
    # Try to find the task at specified location
    return importlib.import_module(find_location)


def get_run_module(task_name):
    return get_mturk_task_module(task_name, '{}.run')


def get_config_module(task_name):
    return get_mturk_task_module(task_name, '{}.task_config')


tornado_settings = {
    "autoescape": None,
    "debug": IS_DEBUG,
    "static_path": get_path('static'),
    "template_path": get_path('static'),
    "compiled_template_cache": False,
}


class PacketWrap(object):
    def __init__(self, data):
        self.data = data
        self.id = None


class Application(tornado.web.Application):
    def __init__(self, port=DEFAULT_PORT, db_file=DEFAULT_DB_FILE, is_sandbox=False):
        self.state = {'is_sandbox': is_sandbox}
        self.subs = {}
        self.sources = {}
        self.port = port
        self.data_handler = MTurkDataHandler(file_name=db_file)
        self.manager = None  # MTurk manager for demo tasks
        self.mturk_manager = MTurkManager.make_taskless_instance(is_sandbox)
        self.mturk_manager.db_logger = self.data_handler
        self.task_manager = None  # This is the task mturk manager

        # TODO load some state from DB

        handlers = [
            (r"/app/(.*)", AppHandler, {'app': self}),
            (r"/run_list", RunListHandler, {'app': self}),
            (r"/task_list", TaskListHandler, {'app': self}),
            (r"/run_task/(.*)", TaskRunHandler, {'app': self}),
            (r"/workers", WorkerListHandler, {'app': self}),
            (r"/runs/(.*)", RunHandler, {'app': self}),
            (r"/workers/(.*)", WorkerHandler, {'app': self}),
            (r"/assignments/(.*)", AssignmentHandler, {'app': self}),
            (r"/approve/(.*)", ApprovalHandler, {'app': self}),
            (r"/reject/(.*)", RejectionHandler, {'app': self}),
            (r"/reverse_rejection/(.*)", ReverseHandler, {'app': self}),
            (r"/block/(.*)", BlockHandler, {'app': self}),
            (r"/bonus/(.*)", BonusHandler, {'app': self}),
            (r"/error/(.*)", ErrorHandler, {'app': self}),
            (r"/socket", TaskSocketHandler, {'app': self}),
            (r"/", RedirectHandler),
        ]
        super(Application, self).__init__(handlers, **tornado_settings)


def broadcast_msg(handler, msg, target_subs=None):
    if target_subs is None:
        target_subs = handler.subs.values()
    for sub in target_subs:
        sub.write_message(json.dumps(msg))


def send_to_sources(handler, msg):
    target_sources = handler.sources.values()
    for source in target_sources:
        source.write_message(json.dumps(msg))


class BaseHandler(tornado.web.RequestHandler):
    def __init__(self, *request, **kwargs):
        self.include_host = False
        super(BaseHandler, self).__init__(*request, **kwargs)

    def write_error(self, status_code, **kwargs):
        logging.error("ERROR: %s: %s" % (status_code, kwargs))
        if self.settings.get("debug") and "exc_info" in kwargs:
            logging.error("rendering error page")
            exc_info = kwargs["exc_info"]
            # exc_info is a tuple consisting of:
            # 1. The class of the Exception
            # 2. The actual Exception that was thrown
            # 3. The traceback opbject
            try:
                params = {
                    'error': exc_info[1],
                    'trace_info': traceback.format_exception(*exc_info),
                    'request': self.request.__dict__,
                }

                self.render("error.html", **params)
                logging.error("rendering complete")
            except Exception as e:
                logging.error(e)


class AppHandler(tornado.web.RequestHandler):
    def initialize(self, app):
        self.state = app.state
        self.subs = app.subs
        self.sources = app.sources
        self.port = app.port
        self.data_handler = app.data_handler

    def get(self, args):
        self.render('index.html', initial_location=args)
        return


class RedirectHandler(tornado.web.RequestHandler):
    def post(self):
        self.set_status(404)

    def get(self):
        print('redirecting')
        self.redirect('/app/home')


class RunListHandler(BaseHandler):
    def initialize(self, app):
        self.state = app.state
        self.subs = app.subs
        self.sources = app.sources
        self.port = app.port
        self.data_handler = app.data_handler

    def post(self):
        req = tornado.escape.json_decode(
            tornado.escape.to_basestring(self.request.body)
        )
        self.write(json.dumps({'t': 'testing!', 'req': req}))

    def get(self):
        results = self.data_handler.get_all_run_data()
        processed_results = []
        for res in results:
            processed_results.append(row_to_dict(res))
        for result in processed_results:
            # TODO implemenent
            result['run_status'] = 'unimplemented'

        self.write(json.dumps(processed_results))


class TaskListHandler(BaseHandler):
    def initialize(self, app):
        self.state = app.state
        self.subs = app.subs
        self.sources = app.sources
        self.port = app.port
        self.data_handler = app.data_handler

    def get(self):
        results = {t: {'task_name': t, 'dir': v} for t, v in tasks.items()}
        for task_name, directory in tasks.items():
            results[task_name]['internal'] = 'parlai_internal' in directory
            results[task_name]['has_custom'] = False
            results[task_name]['react_frontend'] = False
            try:
                config = get_config_module(task_name)
                frontend_version = config.task_config.get('frontend_version')
                if frontend_version is not None and frontend_version >= 1:
                    results[task_name]['react_frontend'] = True
                if os.path.isfile(
                    os.path.join(directory, 'frontend', 'components', 'custom.jsx')
                ):
                    results[task_name]['has_custom'] = True
            except Exception as e:
                print(
                    'Exception {} when loading task details for {}'.format(e, task_name)
                )
                pass
            results[task_name]['active_runs'] = 'unimplemented'  # TODO
            results[task_name]['all_runs'] = 'unimplemented'  # TODO

        self.write(json.dumps(list(results.values())))


def merge_assignments_with_pairings(assignments, pairings, log_id):
    processed_assignments = {}
    for res in assignments:
        assign_dict = row_to_dict(res)
        processed_assignments[assign_dict['assignment_id']] = assign_dict
    for res in pairings:
        pairing_dict = row_to_dict(res)
        assign_id = pairing_dict['assignment_id']
        if assign_id not in processed_assignments:
            print(
                'assignment {} missing from assign table for {}'
                ''.format(assign_id, log_id)
            )
        pairing_dict['world_status'] = pairing_dict['status']
        del pairing_dict['status']
        processed_assignments[assign_id].update(pairing_dict)
    return list(processed_assignments.values())


class RunHandler(BaseHandler):
    def initialize(self, app):
        self.state = app.state
        self.subs = app.subs
        self.sources = app.sources
        self.port = app.port
        self.data_handler = app.data_handler

    def get(self, task_target):
        hits = self.data_handler.get_hits_for_run(task_target)
        processed_hits = []
        for res in hits:
            processed_hits.append(row_to_dict(res))
        assignments = self.data_handler.get_assignments_for_run(task_target)
        pairings = self.data_handler.get_pairings_for_run(task_target)
        processed_assignments = merge_assignments_with_pairings(
            assignments, pairings, 'task {}'.format(task_target)
        )

        workers = set()
        # get feedback data and put into assignments if present
        for assignment in processed_assignments:
            assignment['received_feedback'] = None
            run_id = assignment['run_id']
            conversation_id = assignment['conversation_id']
            worker_id = assignment['worker_id']
            workers.add(worker_id)
            if conversation_id is not None:
                task_data = MTurkDataHandler.get_conversation_data(
                    run_id, conversation_id, worker_id, self.state['is_sandbox']
                )
                if task_data['data'] is not None:
                    assignment['received_feedback'] = task_data['data'].get(
                        'received_feedback'
                    )

        worker_data = {}
        for worker in workers:
            worker_data[worker] = row_to_dict(self.data_handler.get_worker_data(worker))

        run_details = row_to_dict(self.data_handler.get_run_data(task_target))
        # TODO implement run status determination
        run_details['run_status'] = 'unimplemented'
        data = {
            'run_details': run_details,
            'worker_details': worker_data,
            'assignments': processed_assignments,
            'hits': processed_hits,
        }

        self.write(json.dumps(data))


class WorkerListHandler(BaseHandler):
    def initialize(self, app):
        self.state = app.state
        self.subs = app.subs
        self.sources = app.sources
        self.port = app.port
        self.data_handler = app.data_handler

    def get(self):
        results = self.data_handler.get_all_worker_data()
        processed_results = []
        for res in results:
            processed_results.append(row_to_dict(res))

        self.write(json.dumps(processed_results))


class WorkerHandler(BaseHandler):
    def initialize(self, app):
        self.state = app.state
        self.subs = app.subs
        self.sources = app.sources
        self.port = app.port
        self.data_handler = app.data_handler

    def get(self, worker_target):
        assignments = self.data_handler.get_all_assignments_for_worker(worker_target)
        pairings = self.data_handler.get_all_pairings_for_worker(worker_target)
        processed_assignments = merge_assignments_with_pairings(
            assignments, pairings, 'task {}'.format(worker_target)
        )
        worker_details = row_to_dict(self.data_handler.get_worker_data(worker_target))
        data = {'worker_details': worker_details, 'assignments': processed_assignments}

        self.write(json.dumps(data))


class AssignmentHandler(BaseHandler):
    def initialize(self, app):
        self.state = app.state
        self.subs = app.subs
        self.sources = app.sources
        self.port = app.port
        self.data_handler = app.data_handler

    def get(self, assignment_target):
        # Extract assignment
        assignments = [self.data_handler.get_assignment_data(assignment_target)]
        pairings = self.data_handler.get_pairings_for_assignment(assignment_target)
        processed_assignments = merge_assignments_with_pairings(
            assignments, pairings, 'assignment {}'.format(assignment_target)
        )
        assignment = processed_assignments[0]

        # Get assignment details to retrieve assignment content
        run_id = assignment['run_id']
        onboarding_id = assignment['onboarding_id']
        conversation_id = assignment['conversation_id']
        worker_id = assignment['worker_id']

        onboard_data = None
        if onboarding_id is not None:
            onboard_data = MTurkDataHandler.get_conversation_data(
                run_id, onboarding_id, worker_id, self.state['is_sandbox']
            )

        assignment_content = {
            'onboarding': onboard_data,
            'task': MTurkDataHandler.get_conversation_data(
                run_id, conversation_id, worker_id, self.state['is_sandbox']
            ),
            'task_name': '_'.join(run_id.split('_')[:-1]),
        }

        # Get assignment instruction html. This can be much improved
        task_name = '_'.join(run_id.split('_')[:-1])
        try:
            # Try to find the task at specified location
            t = get_config_module(task_name)
            task_instructions = t.task_config['task_description']
        except ImportError:
            task_instructions = None

        data = {
            'assignment_details': assignment,
            'assignment_content': assignment_content,
            'assignment_instructions': task_instructions,
        }

        self.write(json.dumps(data))


class ApprovalHandler(BaseHandler):
    def initialize(self, app):
        self.mturk_manager = app.mturk_manager

    def post(self, assignment_target):
        self.mturk_manager.approve_work(assignment_target)
        data = {'status': True}
        self.write(json.dumps(data))


class ReverseHandler(BaseHandler):
    def initialize(self, app):
        self.mturk_manager = app.mturk_manager

    def post(self, assignment_target):
        self.mturk_manager.approve_work(assignment_target, True)
        print('Assignment {} had rejection reversed'.format(assignment_target))
        data = {'status': True}
        self.write(json.dumps(data))


class RejectionHandler(BaseHandler):
    def initialize(self, app):
        self.mturk_manager = app.mturk_manager

    def post(self, assignment_target):
        data = tornado.escape.json_decode(self.request.body)
        reason = data['reason']
        self.mturk_manager.reject_work(assignment_target, reason)
        print('Rejected {} for reason {}'.format(assignment_target, reason))
        data = {'status': True}
        self.write(json.dumps(data))


class BlockHandler(BaseHandler):
    def initialize(self, app):
        self.mturk_manager = app.mturk_manager

    def post(self, worker_target):
        data = tornado.escape.json_decode(self.request.body)
        reason = data['reason']
        self.mturk_manager.block_worker(worker_target, reason)
        print('Blocked {} for reason {}'.format(worker_target, reason))
        data = {'status': True}
        self.write(json.dumps(data))


class BonusHandler(BaseHandler):
    def initialize(self, app):
        self.mturk_manager = app.mturk_manager

    def post(self, worker_target):
        """
        Requests to /bonus/{worker_id} will give a bonus to that worker.

        Requires a reason, assignment_id, a unique token (for idempotence), and the
        bonus amount IN CENTS
        """
        data = tornado.escape.json_decode(self.request.body)
        reason = data['reason']
        assignment_id = data['assignment_id']
        bonus_cents = data['bonus_cents']
        token = data['bonus_token']

        dollar_amount = bonus_cents / 100.0
        self.mturk_manager.pay_bonus(
            worker_target, dollar_amount, assignment_id, reason, token
        )
        print(
            'Bonused ${} to {} for reason {}'.format(
                dollar_amount, worker_target, reason
            )
        )
        data = {'status': True}
        self.write(json.dumps(data))


class ErrorHandler(BaseHandler):
    def get(self, text):
        error_text = text or "test error"
        raise Exception(error_text)


class TaskSocketHandler(tornado.websocket.WebSocketHandler):
    def initialize(self, app):
        self.app = app
        self.sources = app.sources

    def check_origin(self, origin):
        return True

    def _run_socket(self):
        time.sleep(2)
        asyncio.set_event_loop(asyncio.new_event_loop())
        while self.alive and self.app.task_manager is not None:
            try:
                self.write_message(
                    json.dumps(
                        {
                            'data': [
                                agent.get_update_packet()
                                for agent in self.app.task_manager.agents
                            ],
                            'command': 'sync',
                        }
                    )
                )
                time.sleep(0.2)
            except tornado.websocket.WebSocketClosedError:
                self.alive = False
                self.app.task_manager.timeout_all_agents()

    def open(self):
        self.sid = str(hex(int(time.time() * 10000000))[2:])
        self.alive = True
        if self not in list(self.sources.values()):
            self.sources[self.sid] = self
        logging.info('Opened task socket from ip: {}'.format(self.request.remote_ip))

        self.write_message(json.dumps({'command': 'alive', 'data': 'socket_alive'}))

        t = threading.Thread(target=self._run_socket)
        t.start()

    def on_message(self, message):
        logging.info('from frontend client: {}'.format(message))
        msg = tornado.escape.json_decode(tornado.escape.to_basestring(message))
        message = msg['text']
        task_data = msg['task_data']
        sender_id = msg['sender']
        agent_id = msg['id']
        act = {
            'id': agent_id,
            'task_data': task_data,
            'text': message,
            'message_id': str(uuid.uuid4()),
        }
        t = threading.Thread(
            target=self.app.task_manager.on_new_message,
            args=(sender_id, PacketWrap(act)),
            daemon=True,
        )
        t.start()

    def on_close(self):
        self.alive = False
        if self in list(self.sources.values()):
            self.sources.pop(self.sid, None)


class TaskRunHandler(BaseHandler):
    def initialize(self, app):
        self.app = app

    def post(self, task_target):
        """
        Requests to /run_task/{task_id} will launch a task locally for the given task.

        It will die after 20 mins if it doesn't end on its own.
        """
        try:
            # Load the run and task_config modules from the expected locations
            t = get_run_module(task_target)
            conf = get_config_module(task_target)
            # Set the run file's MTurkManager module to be MockTurkManager
            t.MTurkManager = MockTurkManager
            MockTurkManager.current_manager = None

            # Start a task thread, then wait for the task to be running
            task_thread = threading.Thread(target=t.main, name='Demo-Thread')
            task_thread.start()
            while MockTurkManager.current_manager is None:
                time.sleep(1)
            time.sleep(1)
            manager = MockTurkManager.current_manager

            # Register the current manager, then alive the agents
            self.app.task_manager = manager
            for agent in manager.agents:
                manager.worker_alive(agent.worker_id, agent.hit_id, agent.assignment_id)

            # Tell frontend we're done, and give the initial packets.
            data = {
                'started': True,
                'data': [agent.get_update_packet() for agent in manager.agents],
                'task_config': conf.task_config,
            }
            self.write(json.dumps(data))
        except Exception as e:
            data = {'error': e}
            print(repr(e))
            print(traceback.format_exc())
            self.write(json.dumps(data))


def start_server(
    port=DEFAULT_PORT,
    hostname=DEFAULT_HOSTNAME,
    db_file=DEFAULT_DB_FILE,
    is_sandbox=False,
):
    print("It's Alive!")
    app = Application(port=port, db_file=db_file, is_sandbox=is_sandbox)
    app.listen(port, max_buffer_size=1024 ** 3)
    logging.info("Application Started")

    if "HOSTNAME" in os.environ and hostname == DEFAULT_HOSTNAME:
        hostname = os.environ["HOSTNAME"]
    else:
        hostname = hostname
    print("You can navigate to http://%s:%s" % (hostname, port))
    tornado.ioloop.IOLoop.current().start()


def crawl_dir_for_tasks(search_dir):
    found_dirs = {}  # taskname -> task directory
    contents = os.listdir(search_dir)
    for sub_dir in contents:
        full_sub_dir = os.path.join(search_dir, sub_dir)
        if os.path.exists(os.path.join(full_sub_dir, 'run.py')):
            # assume this is a task
            found_dirs[sub_dir] = full_sub_dir
        if os.path.isdir(full_sub_dir):
            found_dirs.update(crawl_dir_for_tasks(full_sub_dir))
    return found_dirs


def rebuild_source():
    # TODO move task parsing to its own helper file
    # Copy files to the correct locations,
    if os.path.exists('dev/task_components'):
        print('removing old build')
        sh.rm(shlex.split('-rf {}'.format('dev/task_components')))
    os.mkdir('dev/task_components')

    copy_dirs = {}
    parlai_task_dir = os.path.join(parlai_path, 'mturk', 'tasks')
    tasks.update(crawl_dir_for_tasks(parlai_task_dir))

    if parlai_int_path is not None:
        parlai_internal_task_dir = os.path.join(parlai_int_path, 'mturk', 'tasks')
        tasks.update(crawl_dir_for_tasks(parlai_internal_task_dir))

    for task, task_dir in tasks.items():
        if os.path.exists(os.path.join(task_dir, 'frontend')):
            copy_dirs[task] = os.path.join(task_dir, 'frontend')

    for task, task_dir in copy_dirs.items():
        was_built = False
        if 'package.json' in os.listdir(task_dir):
            # need to build this component first
            os.chdir(task_dir)
            packages_installed = subprocess.call(['npm', 'install'])
            if packages_installed != 0:
                raise Exception(
                    'please make sure npm is installed, otherwise '
                    'view the above error for more info.'
                )

            webpack_complete = subprocess.call(['npm', 'run', 'dev'])
            if webpack_complete != 0:
                raise Exception(
                    'Webpack appears to have failed to build your '
                    'frontend. See the above error for more '
                    'information.'
                )
            was_built = True

        os.chdir(here)
        output_dir = os.path.join('dev/task_components', task)
        shutil.copytree(task_dir, output_dir)
        # need to move built custom.jsx back into components
        if was_built:
            shutil.copy2(
                os.path.join(output_dir, 'dist', 'custom.jsx'),
                os.path.join(output_dir, 'components', 'custom.jsx'),
            )
        # Need to give a copy of core_components to have same dir
        shutil.copy2(
            os.path.join(
                parlai_path,
                'mturk',
                'core',
                'react_server',
                'dev',
                'components',
                'core_components.jsx',
            ),
            os.path.join(output_dir, 'components'),
        )

    # Grab up to date version of core components
    shutil.copy2(
        os.path.join(
            parlai_path,
            'mturk',
            'core',
            'react_server',
            'dev',
            'components',
            'core_components.jsx',
        ),
        'dev/task_components',
    )

    # build the full react server
    packages_installed = subprocess.call(['npm', 'install'])
    if packages_installed != 0:
        raise Exception(
            'please make sure npm is installed, otherwise view '
            'the above error for more info.'
        )

    webpack_complete = subprocess.call(['npm', 'run', 'dev'])
    if webpack_complete != 0:
        raise Exception(
            'Webpack appears to have failed to build your '
            'frontend. See the above error for more information.'
        )


def main():
    parser = argparse.ArgumentParser(
        description='Start the ParlAI-MTurk task managing server.'
    )
    parser.add_argument(
        '--port',
        metavar='port',
        type=int,
        default=DEFAULT_PORT,
        help='port to run the server on.',
    )
    parser.add_argument(
        '--hostname',
        metavar='hostname',
        type=str,
        default=DEFAULT_HOSTNAME,
        help='host to run the server on.',
    )
    parser.add_argument(
        '--sandbox',
        dest='sandbox',
        action='store_true',
        default=False,
        help='Run the server using sandbox data',
    )
    parser.add_argument(
        '--db_file',
        metavar='db_file',
        type=str,
        default=DEFAULT_DB_FILE,
        help='name of database to use (in core/run_data)',
    )
    parser.add_argument(
        '--logging_level',
        metavar='logger_level',
        default='INFO',
        help='logging level (default = INFO). Can take logging'
        ' level name or int (example: 20)',
    )
    FLAGS = parser.parse_args()

    if FLAGS.sandbox:
        if FLAGS.db_file == DEFAULT_DB_FILE:
            FLAGS.db_file = DEFAULT_SB_DB_FILE

    logging_level = logging._checkLevel(FLAGS.logging_level)
    logging.getLogger().setLevel(logging_level)

    rebuild_source()

    start_server(
        port=FLAGS.port,
        hostname=FLAGS.hostname,
        db_file=FLAGS.db_file,
        is_sandbox=FLAGS.sandbox,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.mturk.core.worlds import MTurkOnboardWorld, MTurkTaskWorld
import threading


class AskerOnboardingWorld(MTurkOnboardWorld):
    """
    Example onboarding world.

    Sends a message from the world to the worker and then exits as complete after the
    worker uses the interface
    """

    def parley(self):
        ad = {}
        ad['id'] = 'System'
        ad['text'] = (
            "Welcome onboard! You'll be playing the role of the asker. Ask "
            "a question that can be answered with just a number. Send any "
            "message to continue."
        )
        self.mturk_agent.observe(ad)
        self.mturk_agent.act()
        self.episodeDone = True


class AnswererOnboardingWorld(MTurkOnboardWorld):
    """
    Example onboarding world.

    Sends a message from the world to the worker and then exits as complete after the
    worker uses the interface
    """

    def parley(self):
        ad = {}
        ad['id'] = 'System'
        ad['text'] = (
            "Welcome onboard! You'll be playing the role of the answerer. "
            "You'll be asked a question that should be answered with a number. "
            "Answer with something that makes sense. Enter any number to "
            "continue."
        )
        self.mturk_agent.observe(ad)
        self.mturk_agent.act()
        self.episodeDone = True


class EvaluatorOnboardingWorld(MTurkOnboardWorld):
    """
    Example onboarding world.

    Sends a message from the world to the worker and then exits as complete after the
    worker uses the interface
    """

    def parley(self):
        ad = {}
        ad['id'] = 'System'
        ad['text'] = (
            "Welcome onboard! You'll be playing the evaluator. You'll "
            "observe a series of three questions, and then you'll evaluate "
            "whether or not the exchange was accurate. Send an eval to begin."
        )
        self.mturk_agent.observe(ad)
        self.mturk_agent.act()
        self.episodeDone = True


class MultiRoleAgentWorld(MTurkTaskWorld):
    """
    World to demonstrate workers with assymetric roles.

    This task amounts to three rounds and then an evaluation step. It is purposefully
    created as a task to demo multiple views and has no other purpose.
    """

    collector_agent_id = 'Moderator'

    def __init__(self, opt, mturk_agents):
        self.mturk_agents = mturk_agents
        for agent in mturk_agents:
            if agent.demo_role == 'Asker':
                self.asker = agent
            elif agent.demo_role == 'Answerer':
                self.answerer = agent
            else:  # 'Evaluator'
                self.evaluator = agent
        self.episodeDone = False
        self.turns = 0
        self.questions = []
        self.answers = []
        self.accepted = None

    def parley(self):
        if self.turns == 0:
            # Instruction for evaluator
            ad = {'id': 'System', 'text': "Please observe the chat for accuracy."}
            self.evaluator.observe(ad)
        if self.turns < 3:
            # QA pairing
            ad = {
                'id': 'System',
                'text': "Please ask a question with a numeric answer.",
            }
            self.asker.observe(ad)
            question = self.asker.act()
            ad = {'id': 'System', 'text': 'Please answer this question.'}
            self.answerer.observe(ad)
            self.answerer.observe(question)
            self.evaluator.observe(question)
            answer = self.answerer.act()
            self.evaluator.observe(answer)
            self.asker.observe(answer)
            self.questions.append(question)
            self.answers.append(answer)
            self.turns += 1
        else:
            # evaluate
            ad = {'id': 'System', 'text': "Please provide your evaluation."}
            self.evaluator.observe(ad)
            ad = {'id': 'System', 'text': "Please wait for evaluation."}
            self.answerer.observe(ad)
            self.asker.observe(ad)
            self.accepter = self.evaluator.act()
            self.episodeDone = True

    def episode_done(self):
        return self.episodeDone

    def shutdown(self):
        # Parallel shutdown of agents
        def shutdown_agent(agent):
            try:
                agent.shutdown(timeout=None)
            except Exception:
                agent.shutdown()  # not MTurkAgent

        threads = []
        for agent in self.mturk_agents:
            t = threading.Thread(target=shutdown_agent, args=(agent,))
            t.start()
            threads.append(t)
        for t in threads:
            t.join()

    def review_work(self):
        # Can review the work here to accept or reject it
        pass

    def get_custom_task_data(self):
        # brings important data together for the task, to later be used for
        # creating the dataset. If data requires pickling, put it in a field
        # called 'needs-pickle'.
        return {
            'questions': self.questions,
            'answers': self.answers,
            'evaluation': self.accepted,
        }



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import pickle
import time

from botocore.exceptions import ClientError

from parlai.mturk.core.agents import MTurkAgent
from parlai.mturk.core.shared_utils import AssignState
import parlai.mturk.core.data_model as data_model
import parlai.mturk.core.mturk_utils as mturk_utils
import parlai.mturk.core.shared_utils as shared_utils


# Time to persist a disconnect before forgetting about it. Combined with the
# above this will block workers that disconnect at least 10 times in a day
DISCONNECT_PERSIST_LENGTH = 60 * 60 * 24

# Max number of conversation disconnects before a turker should be blocked
MAX_DISCONNECTS = 5

DISCONNECT_FILE_NAME = 'disconnects.pickle'

parent_dir = os.path.dirname(os.path.abspath(__file__))


class WorkerState:
    """
    Class for holding state information about an mturk worker.
    """

    def __init__(self, worker_id, disconnects=0):
        """
        Create a new worker state for the given worker_id.

        Number of prior disconnects is optional.
        """
        self.worker_id = worker_id
        self.agents = {}
        self.disconnects = disconnects

    def active_conversation_count(self):
        """
        Return the number of conversations within this worker state that aren't in a
        final state.
        """
        count = 0
        for assign_id in self.agents:
            if not self.agents[assign_id].is_final():
                count += 1
        return count

    def add_agent(self, mturk_agent):
        """
        Add an assignment to this worker state with the given assign_it.
        """
        assert (
            mturk_agent.worker_id == self.worker_id
        ), "Can't add agent that does not match state's worker_id"
        self.agents[mturk_agent.assignment_id] = mturk_agent

    def get_agent_for_assignment(self, assignment_id):
        return self.agents.get(assignment_id, None)

    def has_assignment(self, assign_id):
        """
        Returns true if this worker has an assignment for the given id.
        """
        return assign_id in self.agents

    def completed_assignments(self):
        """
        Returns the number of assignments this worker has completed.
        """
        complete_count = 0
        for agent in self.agents.values():
            if agent.get_status() == AssignState.STATUS_DONE:
                complete_count += 1
        return complete_count

    def disconnected_assignments(self):
        """
        Returns the number of assignments this worker has completed.
        """
        disconnect_count = 0
        for agent in self.agents.values():
            if agent.get_status() in [
                AssignState.STATUS_DISCONNECT,
                AssignState.STATUS_RETURNED,
            ]:
                disconnect_count += 1
        return disconnect_count


class WorkerManager:
    """
    Class used to keep track of workers state, as well as processing messages that come
    from the web client.
    """

    def __init__(self, mturk_manager, opt):
        self.opt = opt
        self.mturk_manager = mturk_manager
        self.mturk_workers = {}
        self.conv_to_agent = {}
        self.assignment_to_worker_id = {}
        self.hit_id_to_agent = {}
        self.time_blocked_workers = []
        self.load_disconnects()
        self.is_sandbox = mturk_manager.is_sandbox

    def get_worker_data_package(self):
        workers = []
        for mturk_worker in self.mturk_workers.values():
            worker = {
                'id': mturk_worker.worker_id,
                'accepted': len(mturk_worker.agents),
                'completed': mturk_worker.completed_assignments(),
                'disconnected': mturk_worker.disconnected_assignments(),
            }
            workers.append(worker)
        return workers

    def _create_agent(self, hit_id, assignment_id, worker_id):
        """
        Initialize an agent and return it.
        """
        return MTurkAgent(
            self.opt, self.mturk_manager, hit_id, assignment_id, worker_id
        )

    def _get_worker(self, worker_id):
        """
        A safe way to get a worker by worker_id.
        """
        return self.mturk_workers.get(worker_id, None)

    def _get_agent(self, worker_id, assignment_id):
        """
        A safe way to get an agent by worker and assignment_id.
        """
        worker = self._get_worker(worker_id)
        if worker is not None:
            return worker.agents.get(assignment_id, None)
        return None

    def route_packet(self, pkt):
        """
        Put an incoming message into the queue for the agent specified in the packet, as
        they have sent a message from the web client.
        """
        worker_id = pkt.sender_id
        assignment_id = pkt.assignment_id
        agent = self._get_agent(worker_id, assignment_id)
        if agent is not None:
            shared_utils.print_and_log(
                logging.INFO,
                'Manager received: {}'.format(pkt),
                should_print=self.opt['verbose'],
            )
            # Push the message to the message thread to send on a reconnect
            agent.append_message(pkt.data)

            # Clear the send message command, as a message was recieved
            agent.set_last_command(None)
            agent.put_data(pkt.id, pkt.data)

    def map_over_agents(self, map_function, filter_func=None):
        """
        Take an action over all the agents we have access to, filters if a filter_func
        is given.
        """
        for worker in self.mturk_workers.values():
            for agent in worker.agents.values():
                if filter_func is None or filter_func(agent):
                    map_function(agent)

    def get_agent_for_assignment(self, assignment_id):
        """
        Returns agent for the assignment, none if no agent assigned.
        """
        if assignment_id not in self.assignment_to_worker_id:
            return None
        worker_id = self.assignment_to_worker_id[assignment_id]
        worker = self.mturk_workers[worker_id]
        return worker.get_agent_for_assignment(assignment_id)

    def time_block_worker(self, worker_id):
        self.time_blocked_workers.append(worker_id)
        self.mturk_manager.soft_block_worker(worker_id, 'max_time_qual')

    def un_time_block_workers(self, workers=None):
        if workers is None:
            workers = self.time_blocked_workers
            self.time_blocked_workers = []
        for worker_id in workers:
            self.mturk_manager.un_soft_block_worker(worker_id, 'max_time_qual')

    def load_disconnects(self):
        """
        Load disconnects from file, populate the disconnects field for any worker_id
        that has disconnects in the list.

        Any disconnect that occurred longer ago than the disconnect persist length is
        ignored
        """
        self.disconnects = []
        # Load disconnects from file
        file_path = os.path.join(parent_dir, DISCONNECT_FILE_NAME)
        compare_time = time.time()
        if os.path.exists(file_path):
            with open(file_path, 'rb') as f:
                old_disconnects = pickle.load(f)
                self.disconnects = [
                    d
                    for d in old_disconnects
                    if (compare_time - d['time']) < DISCONNECT_PERSIST_LENGTH
                ]
        # Initialize worker states with proper number of disconnects
        for disconnect in self.disconnects:
            worker_id = disconnect['id']
            if worker_id not in self.mturk_workers:
                # add this worker to the worker state
                self.mturk_workers[worker_id] = WorkerState(worker_id)
            self.mturk_workers[worker_id].disconnects += 1

    def save_disconnects(self):
        """
        Saves the local list of disconnects to file.
        """
        file_path = os.path.join(parent_dir, DISCONNECT_FILE_NAME)
        if os.path.exists(file_path):
            os.remove(file_path)
        with open(file_path, 'wb') as f:
            pickle.dump(self.disconnects, f, pickle.HIGHEST_PROTOCOL)

    def handle_agent_disconnect(self, worker_id, assignment_id, partner_callback):
        """
        Handles a disconnect by the given worker, calls partner_callback on all of the
        conversation partners of that worker.
        """
        agent = self._get_agent(worker_id, assignment_id)
        if agent is not None:
            # Disconnect in conversation is not workable
            agent.set_status(AssignState.STATUS_DISCONNECT)
            # in conversation, inform others about disconnect
            conversation_id = agent.conversation_id
            if conversation_id in self.conv_to_agent:
                conv_participants = self.conv_to_agent[conversation_id]
                if agent in conv_participants:
                    for other_agent in conv_participants:
                        if agent.assignment_id != other_agent.assignment_id:
                            partner_callback(other_agent)
                if len(conv_participants) > 1:
                    # The user disconnected from inside a conversation with
                    # another turker, record this as bad behavior
                    self.handle_bad_disconnect(worker_id)

    def handle_bad_disconnect(self, worker_id):
        """
        Update the number of bad disconnects for the given worker, block them if they've
        exceeded the disconnect limit.
        """
        if not self.is_sandbox:
            self.mturk_workers[worker_id].disconnects += 1
            self.disconnects.append({'time': time.time(), 'id': worker_id})
            if self.mturk_workers[worker_id].disconnects > MAX_DISCONNECTS:
                if self.opt['hard_block']:
                    text = (
                        'This worker has repeatedly disconnected from these '
                        'tasks, which require constant connection to complete '
                        'properly as they involve interaction with other '
                        'Turkers. They have been blocked after being warned '
                        'and failing to adhere. This was done in order to '
                        'ensure a better experience for other '
                        'workers who don\'t disconnect.'
                    )
                    self.mturk_manager.block_worker(worker_id, text)
                    shared_utils.print_and_log(
                        logging.INFO,
                        'Worker {} blocked - too many disconnects'.format(worker_id),
                        True,
                    )
                elif self.opt['disconnect_qualification'] is not None:
                    self.mturk_manager.soft_block_worker(
                        worker_id, 'disconnect_qualification'
                    )
                    shared_utils.print_and_log(
                        logging.INFO,
                        'Worker {} soft blocked - too many disconnects'.format(
                            worker_id
                        ),
                        True,
                    )

    def worker_alive(self, worker_id):
        """
        Creates a new worker record if it doesn't exist, returns state.
        """
        if worker_id not in self.mturk_workers:
            self.mturk_workers[worker_id] = WorkerState(worker_id)
        return self.mturk_workers[worker_id]

    def assign_task_to_worker(self, hit_id, assign_id, worker_id):
        self.assignment_to_worker_id[assign_id] = worker_id
        agent = self._create_agent(hit_id, assign_id, worker_id)
        self.hit_id_to_agent[hit_id] = agent
        curr_worker_state = self.mturk_workers[worker_id]
        curr_worker_state.add_agent(agent)

    def get_complete_hits(self):
        """
        Returns the list of all currently completed HITs.
        """
        hit_ids = []
        for hit_id, agent in self.hit_id_to_agent.items():
            if agent.hit_is_complete:
                hit_ids.append(hit_id)
        return hit_ids

    def get_agent_work_status(self, assignment_id):
        """
        Get the current status of an assignment's work.
        """
        client = mturk_utils.get_mturk_client(self.is_sandbox)
        try:
            response = client.get_assignment(AssignmentId=assignment_id)
            status = response['Assignment']['AssignmentStatus']
            worker_id = self.assignment_to_worker_id[assignment_id]
            agent = self._get_agent(worker_id, assignment_id)
            if agent is not None and status == MTurkAgent.ASSIGNMENT_DONE:
                agent.hit_is_complete = True
            return status
        except ClientError as e:
            # If the assignment isn't done, asking for the assignment will fail
            not_done_message = (
                'This operation can be called with a status '
                'of: Reviewable,Approved,Rejected'
            )
            if not_done_message in e.response['Error']['Message']:
                return MTurkAgent.ASSIGNMENT_NOT_DONE
            else:
                shared_utils.print_and_log(
                    logging.WARN,
                    'Unanticipated error in `get_agent_work_status`: '
                    + e.response['Error']['Message'],
                    should_print=True,
                )
                # Assume not done if status check seems to be faulty.
                return MTurkAgent.ASSIGNMENT_NOT_DONE

    def _log_missing_agent(self, worker_id, assignment_id):
        """
        Logs when an agent was expected to exist, yet for some reason it didn't.

        If these happen often there is a problem
        """
        shared_utils.print_and_log(
            logging.WARN,
            'Expected to have an agent for {}_{}, yet none was found'.format(
                worker_id, assignment_id
            ),
        )

    def _get_agent_from_pkt(self, pkt):
        """
        Get the agent object corresponding to this packet's sender.
        """
        worker_id = pkt.sender_id
        assignment_id = pkt.assignment_id
        agent = self._get_agent(worker_id, assignment_id)
        if agent is None:
            self._log_missing_agent(worker_id, assignment_id)
        return agent

    def _change_agent_to_conv(self, pkt):
        """
        Update an agent to a new conversation given a packet from the conversation to be
        switched to.
        """
        agent = self._get_agent_from_pkt(pkt)
        if agent is not None:
            self._assign_agent_to_conversation(agent, agent.conversation_id)

    def _assign_agent_to_conversation(self, agent, conv_id):
        """
        Register an agent object with a conversation id, update status.
        """
        agent.conversation_id = conv_id
        if conv_id not in self.conv_to_agent:
            self.conv_to_agent[conv_id] = []
        self.conv_to_agent[conv_id].append(agent)

    def change_agent_conversation(self, agent, conversation_id, new_agent_id):
        """
        Handle changing a conversation for an agent, takes a callback for when the
        command is acknowledged.
        """
        agent.id = new_agent_id
        agent.conversation_id = conversation_id
        data = {
            'text': data_model.COMMAND_CHANGE_CONVERSATION,
            'conversation_id': conversation_id,
            'agent_id': new_agent_id,
        }
        agent.flush_msg_queue()
        self.mturk_manager.send_command(
            agent.worker_id,
            agent.assignment_id,
            data,
            ack_func=self._change_agent_to_conv,
        )

    def shutdown(self):
        """
        Handles cleaning up and storing state related to workers.
        """
        self.un_time_block_workers()
        self.save_disconnects()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.mturk.core.worlds import MTurkOnboardWorld
from parlai.mturk.core.agents import TIMEOUT_MESSAGE
from parlai.core.worlds import validate, MultiAgentDialogWorld
from joblib import Parallel, delayed
from extract_and_save_personas import main as main_extract
import numpy as np
import time
import os
import pickle
import random

ONBOARD_MSG = '\nWelcome! Below is your persona \
        (you can find it on the left side of the chat)\n \
        When you are ready to start your conversation, \
        click the "I am ready, continue" button below\n'
TIMEOUT_MSG = '<b> The other person has timed out. \
        Please click the "Done with this HIT" button below to finish this HIT.\
        </b>'
WAITING_MSG = 'Please wait while we match you with another worker...'


class PersonasGenerator(object):
    def __init__(self, opt):
        self.personas_idx_stack_path = os.path.join(
            opt['extract_personas_path'], './personas_idx_stack.pkl'
        )

        self.personas_path = '{}/personas-{}'.format(
            opt['extract_personas_path'],
            opt['persona_type'] + 'Revised' if opt['revised'] else 'Original',
        )

        if not os.path.exists(self.personas_path):
            opt['personas_path'] = self.personas_path
            main_extract(opt)
        self.personas_name_list = []

        for f_name in os.listdir(self.personas_path):
            if f_name.endswith('.pkl'):
                self.personas_name_list.append(f_name)

        if os.path.exists(self.personas_idx_stack_path):
            with open(self.personas_idx_stack_path, 'rb') as handle:
                self.idx_stack = pickle.load(handle)
        else:
            self.idx_stack = []
            self.add_idx_stack()
            self.save_idx_stack()
        pass

    def add_idx_stack(self):
        stack = [i for i in range(len(self.personas_name_list))]
        random.seed()
        random.shuffle(stack)
        self.idx_stack = stack + self.idx_stack

    def pop_persona(self):
        if len(self.idx_stack) == 0:
            self.add_idx_stack()
        idx = self.idx_stack.pop()
        data = np.load(
            os.path.join(self.personas_path, self.personas_name_list[int(idx)]),
            allow_pickle=True,
        )
        return (idx, data)

    def push_persona(self, idx):
        self.idx_stack.append(idx)

    def save_idx_stack(self):
        with open(self.personas_idx_stack_path, 'wb') as handle:
            pickle.dump(self.idx_stack, handle)


class PersonaProfileWorld(MTurkOnboardWorld):
    """
    A world that provides a persona to the MTurkAgent.
    """

    def __init__(self, opt, mturk_agent):
        self.task_type = 'sandbox' if opt['is_sandbox'] else 'live'
        self.max_persona_time = opt['max_persona_time']
        super().__init__(opt, mturk_agent)

    def parley(self):
        persona_idx, data = self.mturk_agent.persona_generator.pop_persona()
        self.mturk_agent.persona_idx = persona_idx
        self.mturk_agent.persona_data = data
        persona_text = ''
        for s in data:
            persona_text += '<b><span style="color:blue">' '{}\n</span></b>'.format(
                s.strip()
            )

        self.mturk_agent.observe(
            {
                'id': 'SYSTEM',
                'show_persona': True,
                'text': ONBOARD_MSG + '<br>' + persona_text + '<br>',
            }
        )

        act = self.mturk_agent.act(timeout=self.max_persona_time)

        # timeout
        if act['episode_done'] or (('text' in act and act['text'] == TIMEOUT_MESSAGE)):

            self.mturk_agent.persona_generator.push_persona(
                self.mturk_agent.persona_idx
            )
            self.mturk_agent.persona_generator.save_idx_stack()
            self.episodeDone = True
            return

        if 'text' not in act:
            control_msg = {'id': 'SYSTEM', 'text': WAITING_MSG}
            self.mturk_agent.observe(validate(control_msg))
            self.episodeDone = True


class PersonaChatWorld(MultiAgentDialogWorld):
    def __init__(
        self,
        opt,
        agents=None,
        shared=None,
        range_turn=(4, 7),
        max_turn=10,
        max_resp_time=120,
        world_tag='NONE',
        agent_timeout_shutdown=120,
    ):
        self.agents = agents
        self.turn_idx = 0
        self.range_turn = range_turn
        self.max_turn = max_turn
        self.n_turn = np.random.randint(self.range_turn[0], self.range_turn[1]) + 1
        self.dialog = []
        self.task_type = 'sandbox' if opt['is_sandbox'] else 'live'
        self.chat_done = False
        self.world_tag = world_tag

        # below are timeout protocols
        self.max_resp_time = max_resp_time  # in secs
        self.agent_timeout_shutdown = agent_timeout_shutdown
        super().__init__(opt, agents, shared)

        # get personas
        self.personas = [
            (ag.persona_data if hasattr(ag, 'persona_data') else None)
            for ag in self.agents
        ]

    def parley(self):
        self.turn_idx += 1

        control_msg = {'episode_done': False}
        control_msg['id'] = 'SYSTEM'

        print(self.world_tag + ' is at turn {}...'.format(self.turn_idx))

        """If at first turn, we need to give each agent their persona"""
        if self.turn_idx == 1:
            for idx, agent in enumerate(self.agents):
                persona_text = ''
                for s in self.personas[idx]:
                    persona_text += (
                        '<b><span style="color:blue">'
                        '{}\n</span></b>'.format(s.strip())
                    )
                control_msg['persona_text'] = persona_text
                control_msg['text'] = self.get_instruction(
                    tag='start', agent_id=agent.id
                )
                agent.observe(validate(control_msg))
                if idx == 0:
                    time.sleep(3)

        """If we get to the min turns, inform turker that they can end if they
           want
        """
        if self.turn_idx == self.n_turn + 1:
            for idx, agent in enumerate(self.agents):
                control_msg['text'] = self.get_instruction(idx, tag='exceed_min_turns')
                control_msg['exceed_min_turns'] = True
                agent.observe(validate(control_msg))

        """Otherwise, we proceed accordingly"""
        acts = [None, None]
        for idx, agent in enumerate(self.agents):
            if not self.chat_done:
                acts[idx] = agent.act(timeout=self.max_resp_time)
            if self.check_timeout(acts[idx]):
                return

            if self.turn_idx > 1:
                # only check if message is too short on first message
                while self.is_msg_tooshortlong(acts[idx], agent) or self.is_exact_match(
                    acts[idx], agent
                ):
                    acts[idx] = agent.act()
            else:
                while self.is_exact_match(acts[idx], agent):
                    acts[idx] = agent.act()

            if acts[idx]['episode_done']:
                self.chat_done = True
                for ag in self.agents:
                    # if agent disconnected
                    if ag != agent and ag.some_agent_disconnected:
                        control_msg['text'] = (
                            'The other worker unexpectedly diconnected. '
                            'Please click "Done with this HIT" button below to '
                            'finish this HIT.'
                        )
                        control_msg['episode_done'] = True
                        ag.observe(validate(control_msg))
                        return
                # agent ends chat after exceeding minimum number of turns
                if self.turn_idx > self.n_turn:
                    for ag in self.agents:
                        ag.observe(validate(acts[idx]))
                        control_msg['text'] = (
                            'One of you ended the chat. Thanks for your time! '
                            'Please click "Done with this HIT" button below '
                            'to finish this HIT.'
                        )
                        control_msg['episode_done'] = True
                        ag.observe(validate(control_msg))
                return

            else:
                self.dialog.append((idx, acts[idx]['text']))
                for other_agent in self.agents:
                    if other_agent != agent:
                        other_agent.observe(validate(acts[idx]))

    def shutdown(self):
        global shutdown_agent

        def shutdown_agent(mturk_agent):
            mturk_agent.shutdown()

        Parallel(n_jobs=len(self.agents), backend='threading')(
            delayed(shutdown_agent)(agent) for agent in self.agents
        )

    def episode_done(self):
        return self.chat_done

    def get_instruction(self, agent_id=None, tag='first'):
        if tag == 'start':
            return (
                '\nSuccessfully matched. Now let\'s get to know each other'
                'through the chat! \nYou need to finish at least <b>'
                + str(self.n_turn)
                + ' chat turns</b>, after that you can click the "Done" button '
                'to end the chat. \n'
                '<b>You can track your character description on the left.</b> '
                '\n <span style="color:blue"><b>Please try to speak to the '
                'other person as if you are the character assigned.</b></span>'
                '\n <span style="color:blue"><b>Do not trivially copy the '
                'character descriptions into the message.</b></span>'
            )

        if tag == 'chat_not_done':
            return (
                'Sorry, we need at least <b>'
                + str(self.n_turn + 1 - self.turn_idx)
                + ' more turn(s)</b> to finish. '
                'Please send a new message:'
            )

        if tag == 'timeout':
            return (
                '<b>{}</b> is timeout. Please click the "Done with this HIT" '
                'button below to exit this HIT. No rejections.'.format(agent_id)
            )

        if tag == 'exceed_min_turns':
            return (
                '\n {} chat turns finished! \n Keep chatting or you can click '
                'the "Done" button to end the chat if it\'s your turn.'.format(
                    self.n_turn
                )
            )

    def save_data(self):
        # save persona_idx_stack
        convo_finished = True
        bad_workers = []
        for ag in self.agents:
            if (
                ag.hit_is_abandoned
                or ag.hit_is_returned
                or ag.disconnected
                or ag.hit_is_expired
            ):
                bad_workers.append(ag.worker_id)
                convo_finished = False
        if not convo_finished or self.dialog == []:
            for ag in self.agents:
                ag.not_approve = True
                ag.persona_generator.push_persona(ag.persona_idx)
                print(
                    "\n******* Push persona {} back to stack. *******\n".format(
                        ag.persona_idx
                    )
                )

        data_path = self.opt['extract_personas_path']
        if not os.path.exists(data_path):
            os.makedirs(data_path)
        if convo_finished:
            filename = os.path.join(
                data_path,
                '{}_{}_{}.pkl'.format(
                    time.strftime("%Y%m%d-%H%M%S"),
                    np.random.randint(0, 1000),
                    self.task_type,
                ),
            )
        else:
            filename = os.path.join(
                data_path,
                '{}_{}_{}_incomplete.pkl'.format(
                    time.strftime("%Y%m%d-%H%M%S"),
                    np.random.randint(0, 1000),
                    self.task_type,
                ),
            )
        print(self.world_tag + ': Data successfully saved at {}.'.format(filename))
        pickle.dump(
            {
                'personas': self.personas,
                'dialog': self.dialog,
                'workers': [ag.worker_id for ag in self.agents],
                'bad_workers': bad_workers,
                'n_turn': self.n_turn,
            },
            open(filename, 'wb'),
        )

    def is_exact_match(self, act, ag, tolerance=0):
        if act['episode_done']:
            return False

        control_msg = {'episode_done': False}
        control_msg['id'] = 'SYSTEM'

        text = act['text']
        if text not in ['', ' ', '  ', '   ']:
            n_word_match = 0
            for per in ag.persona_data:
                per_parse = per.split(' ')
                regular_words = ['', ' ', 'I', 'I\'m', 'My', 'i']
                for r_w in regular_words:
                    if r_w in per_parse:
                        per_parse.remove(r_w)
                per_subseq = [
                    ' '.join(per_parse[i : i + len(per_parse) - tolerance])
                    for i in range(tolerance + 1)
                ]
                for pp in per_subseq:
                    if pp in ['', ' ', '  ', '   ']:
                        per_subseq.remove(pp)
                n_word_match += sum([(paa in text) for paa in per_subseq])
            if n_word_match > 0:
                control_msg['text'] = (
                    'We found that you <b><span style="color:red">trivially '
                    'copied character descriptions</span></b>. Please '
                    'rephrase your message again.'
                )
                ag.observe(validate(control_msg))
                return True
            else:
                return False

    def is_msg_tooshortlong(self, act, ag, th_min=5, th_max=17):
        if act['episode_done']:
            return False

        control_msg = {'episode_done': False}
        control_msg['id'] = 'SYSTEM'

        msg_len = len(act['text'].split(' '))
        if msg_len < th_min:
            control_msg['text'] = (
                'Your message is too short, please make it more than '
                '<b><span style="color:red">5 words</span></b>.'
            )
            ag.observe(validate(control_msg))
            return True
        if msg_len > th_max:
            control_msg['text'] = (
                'Your message is too long, please make it less than '
                '<b><span style="color:red">15 words</span></b>.'
            )
            ag.observe(validate(control_msg))
            return True
        return False

    def reset_random(self):
        self.n_turn = np.random.randint(self.range_turn[0], self.range_turn[1]) + 1

    def check_timeout(self, act):
        if act['text'] == '[TIMEOUT]' and act['episode_done']:
            control_msg = {'episode_done': True}
            control_msg['id'] = 'SYSTEM'
            control_msg['text'] = self.get_instruction(
                agent_id=act['id'], tag='timeout'
            )
            for ag in self.agents:
                if ag.id != act['id']:
                    ag.observe(validate(control_msg))
            self.chat_done = True
            return True
        else:
            return False

    def review_work(self):
        global review_agent

        def review_agent(ag):
            if hasattr(ag, 'not_approve'):
                pass
            else:
                ag.approve_work()

        Parallel(n_jobs=len(self.agents), backend='threading')(
            delayed(review_agent)(agent) for agent in self.agents
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

task_config = {}

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config['hit_title'] = 'Rehprase a Character Description'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config[
    'hit_description'
] = 'You will rephrase some character (persona) descriptions'


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat,dialog,text,game'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config[
    'task_description'
] = '''
In this task, we will show you 4 to 5 sentences with each of them describes
some person's characteristics.
<br>
Your jobs is to <b><span style="color:blue">rephrase the sentence to a new one
which is about a relative characteristic that the same person may
have.</span></b>
<br><br>
<b><span style="color:red">Do not trivially rephrase by copying the words in
the original sentences. Make a natural rephrasing.</span></b>
See examples below, BAD EXAMPLES have trivial word matching and weird
rephrasing.
<br>
<br>
<b><span style="color:blue">GOOD EXAMPLE</span></b>: rephrase "<b>My father
worked for Ford.</b>" to
"<b><span style="color:blue">My dad worked in the car industry.</span></b>"
<br>
<b><span style="color:red">BAD EXAMPLE</span></b>: rephrase "<b>My father
worked for Ford.</b>" to
"<b><span style="color:red">My dad was employed by Ford.</span></b>"
(trivial word matching like "Ford")
<br>
<br>
<b><span style="color:blue">GOOD EXAMPLE:</span></b> rephrase
"<b>I like basketball.</b>" to
"<b><span style="color:blue">I am big fan of Michael Jordan.</span></b>"
<br>
<b><span style="color:red">BAD EXAMPLE</span></b>: rephrase "<b>I like
basketball.</b>" to
"<b><span style="color:red">I am good at basketball.</span></b>"
(trivial word matching like "basketball")
<br>
<br>
<b><span style="color:blue">GOOD EXAMPLE:</span></b> rephrase "<b>I cannot
choose between lollipops and rainbows</b>" to
"<b><span style="color:blue">I like candies.</span></b>"
<br>
<b><span style="color:red">BAD EXAMPLE</span></b>: rephrase "<b>I cannot choose
between lollipops and rainbows</b>" to
"<b><span style="color:red">Suckers and brightly colored prism reflections are
two of my favorites</span></b>" (unnatural phrases like "prism reflections")
<br>
<br>
<b><span style="color:blue">GOOD EXAMPLE:</span></b> rephrase "<b>I like eating
pretzels</b>" to
"<b><span style="color:blue">I enjoy beers and beer snacks.</span></b>"
<br>
<b><span style="color:red">BAD EXAMPLE</span></b>: rephrase "<b>I like eating
pretzels</b>" to
"<b><span style="color:red">I like to chew and swallow twisted bread with
salt</span></b>" (unnatural description)
<br>
<br>
After you finish, click “Done with this HIT” to submit.
<br>
<br>
- Do not reference the task or MTurk itself when rephrasing the character.
<br>
- No racism, sexism or otherwise offensive comments, or the submission will be rejected.
'''



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import errno
import logging
import json
import threading
import time
from queue import PriorityQueue, Empty
import websocket

from parlai.mturk.core.shared_utils import print_and_log
import parlai.mturk.core.data_model as data_model
import parlai.mturk.core.shared_utils as shared_utils


class Packet:
    """
    Class for holding information sent over a socket.
    """

    # Possible Packet Status
    STATUS_NONE = -1
    STATUS_INIT = 0
    STATUS_SENT = 1
    STATUS_ACK = 2
    STATUS_FAIL = 3

    # Possible Packet Types
    TYPE_ACK = 'ack'
    TYPE_ALIVE = 'alive'
    TYPE_MESSAGE = 'message'
    TYPE_HEARTBEAT = 'heartbeat'
    TYPE_PONG = 'pong'

    def __init__(
        self,
        id,
        type,
        sender_id,
        receiver_id,
        assignment_id,
        data,
        conversation_id=None,
        requires_ack=None,
        blocking=None,
        ack_func=None,
    ):
        """
        Create a packet to be used for holding information before it is
        sent through the socket
        id:               Unique ID to distinguish this packet from others
        type:             TYPE of packet (ACK, ALIVE, MESSAGE, HEARTBEAT)
        sender_id:        Sender ID for this packet
        receiver_id:      Recipient ID for this packet
        assignment_id:    Assignment ID for this packet
        data:             Contents of the packet
        conversation_id:  Packet metadata - what conversation this belongs to
        requires_ack:     Whether or not this packet needs to be acknowledged,
                           determines if retry logic will be used until ack is
                           recieved. Default for messages and alives is True,
                           default for the rest is false.
        blocking:         Whether or not this packet requires blocking to
                           remain in order amongst other packets in the queue.
                           Default is for messages and alives is True, default
                           for the rest is false.
        ack_func:         Function to call upon successful ack of a packet
                           Default calls no function on ack
        """
        self.id = id
        self.type = type
        self.sender_id = sender_id
        self.receiver_id = receiver_id
        self.assignment_id = assignment_id
        self.data = data
        self.conversation_id = conversation_id
        important = self.type in [self.TYPE_ALIVE, self.TYPE_MESSAGE]
        self.requires_ack = important if requires_ack is None else requires_ack
        self.blocking = important if blocking is None else blocking
        self.ack_func = ack_func
        self.status = self.STATUS_INIT

    @staticmethod
    def from_dict(packet):
        """
        Create a packet from the dictionary that would be recieved over a socket.
        """
        try:
            packet_id = packet['id']
            packet_type = packet['type']
            sender_id = packet['sender_id']
            receiver_id = packet['receiver_id']
            assignment_id = packet['assignment_id']
            data = None
            if 'data' in packet:
                data = packet['data']
            else:
                data = ''
            conversation_id = packet['conversation_id']

            return Packet(
                packet_id,
                packet_type,
                sender_id,
                receiver_id,
                assignment_id,
                data,
                conversation_id,
            )
        except Exception:
            print_and_log(
                logging.WARN,
                'Could not create a valid packet out of the dictionary'
                'provided: {}'.format(packet),
            )
            return None

    def as_dict(self):
        """
        Convert a packet into a form that can be pushed over a socket.
        """
        return {
            'id': self.id,
            'type': self.type,
            'sender_id': self.sender_id,
            'receiver_id': self.receiver_id,
            'assignment_id': self.assignment_id,
            'conversation_id': self.conversation_id,
            'data': self.data,
        }

    def get_sender_connection_id(self):
        """
        Get the connection_id that this packet came from.
        """
        return '{}_{}'.format(self.sender_id, self.assignment_id)

    def get_receiver_connection_id(self):
        """
        Get the connection_id that this is going to.
        """
        return '{}_{}'.format(self.receiver_id, self.assignment_id)

    def get_ack(self):
        """
        Return a new packet that can be used to acknowledge this packet.
        """
        return Packet(
            self.id,
            self.TYPE_ACK,
            self.receiver_id,
            self.sender_id,
            self.assignment_id,
            '',
            self.conversation_id,
            False,
            False,
        )

    def new_copy(self):
        """
        Return a new packet that is a copy of this packet with a new id and with a fresh
        status.
        """
        packet = Packet.from_dict(self.as_dict())
        packet.id = shared_utils.generate_event_id(self.receiver_id)
        packet.blocking = self.blocking
        packet.requires_ack = self.requires_ack
        return packet

    def __repr__(self):
        return 'Packet <{}>'.format(self.as_dict())

    def swap_sender(self):
        """
        Swaps the sender_id and receiver_id.
        """
        self.sender_id, self.receiver_id = self.receiver_id, self.sender_id
        return self

    def set_type(self, new_type):
        """
        Updates the message type.
        """
        self.type = new_type
        return self

    def set_data(self, new_data):
        """
        Updates the message data.
        """
        self.data = new_data
        return self


class SocketManager:
    """
    SocketManager is a wrapper around websocket to stabilize its packet passing.

    The manager handles resending packet, as well as maintaining alive status for all
    the connections it forms
    """

    # Time to acknowledge different message types
    ACK_TIME = {Packet.TYPE_ALIVE: 4, Packet.TYPE_MESSAGE: 4}

    # Default pongs without heartbeat before socket considered dead
    DEF_MISSED_PONGS = 20
    HEARTBEAT_RATE = 4
    DEF_DEAD_TIME = 30

    def __init__(
        self,
        server_url,
        port,
        alive_callback,
        message_callback,
        socket_dead_callback,
        task_group_id,
        socket_dead_timeout=None,
        server_death_callback=None,
    ):
        """
        server_url:           url at which the server is to be run
        port:                 port for the socket to operate on
        alive_callback:       function to be called on alive Packets, defined
                               alive_callback(self, pkt)
        message_callback:     function to be called on message Packets, defined
                               message_callback(self, pkt)
        socket_dead_callback: function to be called when a socket dies, should
                              return false if the socket_manager should ignore
                              the death and treat the socket as alive defined
                               on_socket_dead(self, worker_id, assignment_id)
        socket_dead_timeout:  time to wait between heartbeats before dying
        """
        self.server_url = server_url
        self.port = port
        self.alive_callback = alive_callback
        self.message_callback = message_callback
        self.socket_dead_callback = socket_dead_callback
        self.server_death_callback = server_death_callback
        if socket_dead_timeout is not None:
            self.missed_pongs = 1 + socket_dead_timeout / self.HEARTBEAT_RATE
        else:
            self.missed_pongs = self.DEF_MISSED_PONGS
        self.task_group_id = task_group_id

        self.ws = None
        self.keep_running = True

        # initialize the state
        self.listen_thread = None
        self.send_thread = None
        self.queues = {}
        self.blocking_packets = {}  # connection_id to blocking packet map
        self.threads = {}
        self.run = {}
        self.last_sent_heartbeat_time = {}  # time of last heartbeat sent
        self.last_received_heartbeat = {}  # actual last received heartbeat
        self.pongs_without_heartbeat = {}
        # TODO update processed packets to *ensure* only one execution per
        # packet, as right now two threads can be spawned to process the
        # same packet at the same time, as the packet is only added to this
        # set after processing is complete.
        self.processed_packets = set()
        self.packet_map = {}
        self.alive = False
        self.is_shutdown = False
        self.send_lock = threading.Condition()
        self.packet_map_lock = threading.Condition()
        self.thread_shutdown_lock = threading.Condition()
        self.worker_assign_ids = {}  # mapping from connection id to pair

        # setup the socket
        self._setup_socket()

    def get_my_sender_id(self):
        """
        Gives the name that this socket manager should use for its world.
        """
        return '[World_{}]'.format(self.task_group_id)

    def _safe_send(self, data, force=False):
        if not self.alive and not force:
            # Try to wait a half second to send a packet
            timeout = 0.5
            while timeout > 0 and not self.alive:
                time.sleep(0.1)
                timeout -= 0.1
            if not self.alive:
                # don't try to send a packet if we're still dead
                return False
        try:
            with self.send_lock:
                self.ws.send(data)
        except websocket.WebSocketConnectionClosedException:
            # The channel died mid-send, wait for it to come back up
            return False
        except BrokenPipeError:  # noqa F821 we don't support p2
            # The channel died mid-send, wait for it to come back up
            return False
        except AttributeError:
            # _ensure_closed was called in parallel, self.ws = None
            return False
        except Exception as e:
            shared_utils.print_and_log(
                logging.WARN,
                'Unexpected socket error occured: {}'.format(repr(e)),
                should_print=True,
            )
            return False
        return True

    def _ensure_closed(self):
        self.alive = False
        if self.ws is None:
            return
        try:
            self.ws.close()
        except websocket.WebSocketConnectionClosedException:
            pass
        self.ws = None

    def _send_world_alive(self):
        """
        Registers world with the passthrough server.
        """
        self._safe_send(
            json.dumps(
                {
                    'type': data_model.SOCKET_AGENT_ALIVE_STRING,
                    'content': {
                        'id': 'WORLD_ALIVE',
                        'sender_id': self.get_my_sender_id(),
                    },
                }
            ),
            force=True,
        )

    def _send_needed_heartbeat(self, connection_id):
        """
        Sends a heartbeat to a connection if needed.
        """
        if connection_id not in self.last_received_heartbeat:
            return
        if self.last_received_heartbeat[connection_id] is None:
            return
        if (
            time.time() - self.last_sent_heartbeat_time[connection_id]
            < self.HEARTBEAT_RATE
        ):
            return
        packet = self.last_received_heartbeat[connection_id]
        self._safe_send(
            json.dumps(
                {
                    'type': data_model.SOCKET_ROUTE_PACKET_STRING,
                    'content': packet.new_copy().swap_sender().set_data('').as_dict(),
                }
            )
        )
        self.last_sent_heartbeat_time[connection_id] = time.time()

    def _send_ack(self, packet):
        """
        Sends an ack to a given packet.
        """
        ack = packet.get_ack().as_dict()
        result = self._safe_send(
            json.dumps({'type': data_model.SOCKET_ROUTE_PACKET_STRING, 'content': ack})
        )
        if result:
            packet.status = Packet.STATUS_SENT

    def _send_packet(self, packet, connection_id, send_time):
        """
        Sends a packet, blocks if the packet is blocking.
        """
        # Send the packet
        pkt = packet.as_dict()
        if pkt['data'] is None or packet.status == Packet.STATUS_ACK:
            return  # This packet was _just_ acked.
        shared_utils.print_and_log(logging.DEBUG, 'Send packet: {}'.format(packet))

        result = self._safe_send(
            json.dumps({'type': data_model.SOCKET_ROUTE_PACKET_STRING, 'content': pkt})
        )
        if not result:
            # The channel died mid-send, wait for it to come back up
            self._safe_put(connection_id, (send_time, packet))
            return

        if packet.status != Packet.STATUS_ACK:
            packet.status = Packet.STATUS_SENT

        # Handles acks and blocking
        if packet.requires_ack:
            if packet.blocking:
                # Put the packet right back into its place to prevent sending
                # other packets, then block that connection
                self._safe_put(connection_id, (send_time, packet))
                t = time.time() + self.ACK_TIME[packet.type]
                self.blocking_packets[connection_id] = (t, packet)
            else:
                # non-blocking ack: add ack-check to queue
                t = time.time() + self.ACK_TIME[packet.type]
                self._safe_put(connection_id, (t, packet))

    def _spawn_reaper_thread(self):
        def _reaper_thread(*args):
            start_time = time.time()
            wait_time = self.DEF_MISSED_PONGS * self.HEARTBEAT_RATE
            while time.time() - start_time < wait_time:
                if self.is_shutdown:
                    return
                if self.alive:
                    return
                time.sleep(0.3)
            if self.server_death_callback is not None:
                shared_utils.print_and_log(
                    logging.WARN,
                    'Server has disconnected and could not reconnect. '
                    'Assuming the worst and calling the death callback. '
                    '(Usually shutdown)',
                    should_print=True,
                )
                self.server_death_callback()

        reaper_thread = threading.Thread(
            target=_reaper_thread, name='socket-reaper-{}'.format(self.task_group_id)
        )
        reaper_thread.daemon = True
        reaper_thread.start()

    def _setup_socket(self):
        """
        Create socket handlers and registers the socket.
        """

        def on_socket_open(*args):
            shared_utils.print_and_log(logging.DEBUG, 'Socket open: {}'.format(args))
            self._send_world_alive()

        def on_error(ws, error):
            try:
                if error.errno == errno.ECONNREFUSED:
                    self.use_socket = False
                    self._ensure_closed()
                    raise Exception("Socket refused connection, cancelling")
                else:
                    shared_utils.print_and_log(
                        logging.WARN, 'Socket logged error: {}'.format(error)
                    )
                    self._ensure_closed()
            except Exception:
                if type(error) is websocket.WebSocketConnectionClosedException:
                    return  # Connection closed is noop
                shared_utils.print_and_log(
                    logging.WARN,
                    'Socket logged error: {} Restarting'.format(repr(error)),
                )
                self._ensure_closed()

        def on_disconnect(*args):
            """
            Disconnect event is a no-op for us, as the server reconnects automatically
            on a retry.

            Just in case the server is actually dead we set up a thread to reap the
            whole task.
            """
            shared_utils.print_and_log(
                logging.INFO, 'World server disconnected: {}'.format(args)
            )
            self._ensure_closed()
            if not self.is_shutdown:
                self._spawn_reaper_thread()

        def on_message(*args):
            """
            Incoming message handler for ACKs, ALIVEs, HEARTBEATs, PONGs, and MESSAGEs.
            """
            packet_dict = json.loads(args[1])
            if packet_dict['type'] == 'conn_success':
                self.alive = True
                return  # No action for successful connection
            packet = Packet.from_dict(packet_dict['content'])
            if packet is None:
                return
            packet_id = packet.id
            packet_type = packet.type
            connection_id = packet.get_sender_connection_id()
            if packet_type == Packet.TYPE_ACK:
                with self.packet_map_lock:
                    if packet_id not in self.packet_map:
                        # Don't do anything when acking a packet we don't have
                        return
                    # Acknowledgements should mark a packet as acknowledged
                    shared_utils.print_and_log(
                        logging.DEBUG, 'On new ack: {}'.format(args)
                    )
                    self.packet_map[packet_id].status = Packet.STATUS_ACK
                    # If the packet sender wanted to do something on ack
                    if self.packet_map[packet_id].ack_func:
                        self.packet_map[packet_id].ack_func(packet)
                    # clear the stored packet data for memory reasons
                    try:
                        self.packet_map[packet_id].data = None
                    except Exception:
                        pass  # state already reduced, perhaps by ack_func
            elif packet_type == Packet.TYPE_HEARTBEAT:
                # Heartbeats update the last heartbeat, clears pongs w/o beat
                self.last_received_heartbeat[connection_id] = packet
                self.pongs_without_heartbeat[connection_id] = 0
            elif packet_type == Packet.TYPE_PONG:
                # Message in response from the router, ensuring we're connected
                # to it. Redundant but useful for metering from web client.
                pong_conn_id = packet.get_receiver_connection_id()
                if self.last_received_heartbeat[pong_conn_id] is not None:
                    self.pongs_without_heartbeat[pong_conn_id] += 1
            elif packet_id in self.processed_packets:
                # We don't want to re-process a packet that has already
                # been recieved, but we do need to tell the sender it is ack'd
                # re-acknowledge the packet was recieved and already processed
                self._send_ack(packet)
            else:

                # Remaining packet types need to be acknowledged
                shared_utils.print_and_log(
                    logging.DEBUG, 'On new message: {}'.format(args)
                )
                # Call the appropriate callback
                if packet_type == Packet.TYPE_ALIVE:
                    self.alive_callback(packet)
                elif packet_type == Packet.TYPE_MESSAGE:
                    self.message_callback(packet)

                # acknowledge the packet was recieved and processed
                self._send_ack(packet)

                # Note to self that this packet has already been processed,
                # and shouldn't be processed again in the future
                self.processed_packets.add(packet_id)

        def run_socket(*args):
            url_base_name = self.server_url.split('https://')[1]
            protocol = "wss"
            if url_base_name in ['localhost', '127.0.0.1']:
                protocol = "ws"
            while self.keep_running:
                try:
                    sock_addr = "{}://{}:{}/".format(protocol, url_base_name, self.port)
                    self.ws = websocket.WebSocketApp(
                        sock_addr,
                        on_message=on_message,
                        on_error=on_error,
                        on_close=on_disconnect,
                    )
                    self.ws.on_open = on_socket_open
                    self.ws.run_forever(ping_interval=8 * self.HEARTBEAT_RATE)
                    self._ensure_closed()
                except Exception as e:
                    shared_utils.print_and_log(
                        logging.WARN,
                        'Socket error {}, attempting restart'.format(repr(e)),
                    )
                time.sleep(0.2)

        # Start listening thread
        self.listen_thread = threading.Thread(
            target=run_socket, name='Main-Socket-Recv-Thread'
        )
        self.listen_thread.daemon = True
        self.listen_thread.start()
        time.sleep(1.2)
        start_time = time.time()
        while not self.alive:
            if time.time() - start_time > self.DEF_DEAD_TIME:
                self.server_death_callback()
                raise ConnectionRefusedError(  # noqa F821 we only support py3
                    'Was not able to establish a connection with the server, '
                    'please try to run again. If that fails,'
                    'please ensure that your local device has the correct SSL '
                    'certs installed.'
                )
            try:
                self._send_world_alive()
            except Exception:
                pass
            time.sleep(self.HEARTBEAT_RATE / 2)

        # Start sending thread
        self.send_thread = threading.Thread(
            target=self.channel_thread, name='Main-Socket-Send-Thread'
        )
        self.send_thread.daemon = True
        self.send_thread.start()

    def packet_should_block(self, packet_item):
        """
        Helper function to determine if a packet is still blocking.
        """
        t, packet = packet_item
        if time.time() > t:
            return False  # Exceeded blocking time
        if packet.status in [Packet.STATUS_ACK, Packet.STATUS_FAIL]:
            return False  # No longer in blocking status
        return True

    def channel_thread(self):
        """
        Handler thread for monitoring all channels.
        """
        # while the thread is still alive
        while not self.is_shutdown:
            if self.ws is None:
                # Waiting for websocket to come back alive
                time.sleep(shared_utils.THREAD_SHORT_SLEEP)
                continue
            for connection_id in self.run.copy():
                if not self.run[connection_id]:
                    continue
                try:
                    # Send a heartbeat if needed
                    self._send_needed_heartbeat(connection_id)
                    # Check if client is still alive
                    if self.pongs_without_heartbeat[connection_id] > self.missed_pongs:
                        self.run[connection_id] = False
                        # Setup and run the channel sending thread
                        self.threads[connection_id] = threading.Thread(
                            target=self.socket_dead_callback,
                            name='Socket-dead-{}'.format(connection_id),
                            args=self.worker_assign_ids[connection_id],
                        )
                        self.threads[connection_id].daemon = True
                        self.threads[connection_id].start()

                    # Make sure the queue still exists
                    if connection_id not in self.queues:
                        self.run[connection_id] = False
                        break
                    if self.blocking_packets.get(connection_id) is not None:
                        packet_item = self.blocking_packets[connection_id]
                        if not self.packet_should_block(packet_item):
                            self.blocking_packets[connection_id] = None
                        else:
                            continue
                    try:
                        # Get first item in the queue, check if can send it yet
                        item = self.queues[connection_id].get(block=False)
                        t = item[0]
                        if time.time() < t:
                            # Put the item back into the queue,
                            # it's not time to pop yet
                            self._safe_put(connection_id, item)
                        else:
                            # Try to send the packet
                            packet = item[1]
                            if not packet:
                                # This packet was deleted out from under us
                                continue
                            if packet.status is not Packet.STATUS_ACK:
                                # either need to send initial packet
                                # or resend not-acked packet
                                self._send_packet(packet, connection_id, t)
                    except Empty:
                        pass
                except Exception as e:
                    shared_utils.print_and_log(
                        logging.WARN,
                        'Unexpected error occurred in socket handling thread: '
                        '{}'.format(repr(e)),
                        should_print=True,
                    )
            time.sleep(shared_utils.THREAD_SHORT_SLEEP)

    def open_channel(self, worker_id, assignment_id):
        """
        Opens a channel for a worker on a given assignment, doesn't re-open if the
        channel is already open.
        """
        connection_id = '{}_{}'.format(worker_id, assignment_id)
        if connection_id in self.queues and self.run[connection_id]:
            shared_utils.print_and_log(
                logging.DEBUG, 'Channel ({}) already open'.format(connection_id)
            )
            return
        self.run[connection_id] = True
        self.queues[connection_id] = PriorityQueue()
        self.last_sent_heartbeat_time[connection_id] = 0
        self.pongs_without_heartbeat[connection_id] = 0
        self.last_received_heartbeat[connection_id] = None
        self.worker_assign_ids[connection_id] = (worker_id, assignment_id)

    def close_channel(self, connection_id):
        """
        Closes a channel by connection_id.
        """
        shared_utils.print_and_log(
            logging.DEBUG, 'Closing channel {}'.format(connection_id)
        )
        self.run[connection_id] = False
        with self.thread_shutdown_lock:
            if connection_id in self.queues:
                while not self.queues[connection_id].empty():  # Empty queue
                    try:
                        item = self.queues[connection_id].get(block=False)
                        t = item[0]
                        packet = item[1]
                        if not packet:
                            continue
                        packet.requires_ack = False
                        packet.blocking = False
                        if packet.status is not Packet.STATUS_ACK:
                            self._send_packet(packet, connection_id, t)
                    except Empty:
                        break
                # Clean up packets
                with self.packet_map_lock:
                    packet_ids = list(self.packet_map.keys())
                    for packet_id in packet_ids:
                        packet = self.packet_map[packet_id]
                        packet_conn_id = packet.get_receiver_connection_id()
                        if connection_id == packet_conn_id:
                            del self.packet_map[packet_id]
                # Clean up other resources
                del self.queues[connection_id]

    def close_all_channels(self):
        """
        Closes all channels by clearing the list of channels.
        """
        shared_utils.print_and_log(logging.DEBUG, 'Closing all channels')
        connection_ids = list(self.queues.keys())
        for connection_id in connection_ids:
            self.close_channel(connection_id)

    def socket_is_open(self, connection_id):
        return connection_id in self.queues

    def queue_packet(self, packet):
        """
        Queues sending a packet to its intended owner, returns True if queued
        successfully and False if there is no such worker (anymore)
        """
        connection_id = packet.get_receiver_connection_id()
        if not self.socket_is_open(connection_id):
            # Warn if there is no socket to send through for the expected recip
            shared_utils.print_and_log(
                logging.WARN,
                'Can not send packet to worker_id {}: packet queue not found. '
                'Message: {}'.format(connection_id, packet.data),
            )
            return False
        shared_utils.print_and_log(
            logging.DEBUG,
            'Put packet ({}) in queue ({})'.format(packet.id, connection_id),
        )
        # Get the current time to put packet into the priority queue
        with self.packet_map_lock:
            self.packet_map[packet.id] = packet
        item = (time.time(), packet)
        self._safe_put(connection_id, item)
        return True

    def get_status(self, packet_id):
        """
        Returns the status of a particular packet by id.
        """
        with self.packet_map_lock:
            if packet_id not in self.packet_map:
                return Packet.STATUS_NONE
            return self.packet_map[packet_id].status

    def _safe_put(self, connection_id, item):
        """
        Ensures that a queue exists before putting an item into it, logs if there's a
        failure.
        """
        if connection_id in self.queues:
            self.queues[connection_id].put(item)
        else:
            item[1].status = Packet.STATUS_FAIL
            shared_utils.print_and_log(
                logging.WARN,
                'Queue {} did not exist to put a message in'.format(connection_id),
            )

    def shutdown(self):
        """
        marks the socket manager as closing, shuts down all channels.
        """
        self.is_shutdown = True
        self.close_all_channels()
        self.keep_running = False
        self._ensure_closed()


# TODO extract common functionality from the above and make both classes extend
# from the base
class StaticSocketManager(SocketManager):
    """
    Version of SocketManager that communicates consistently with the world, but isn't
    keeping track of the liveliness of the agents that connect as these are single
    person tasks.

    Submissions are handled via post rather than served over socket, so it doesn't make
    sense to.
    """

    def channel_thread(self):
        """
        Handler thread for monitoring all channels to send things to.
        """
        # while the thread is still alive
        while not self.is_shutdown:
            for connection_id in self.run.copy():
                if not self.run[connection_id]:
                    continue
                try:
                    # Make sure the queue still exists
                    if connection_id not in self.queues:
                        self.run[connection_id] = False
                        break
                    if self.blocking_packets.get(connection_id) is not None:
                        packet_item = self.blocking_packets[connection_id]
                        if not self.packet_should_block(packet_item):
                            self.blocking_packets[connection_id] = None
                        else:
                            continue
                    try:
                        # Get first item in the queue, check if can send it yet
                        item = self.queues[connection_id].get(block=False)
                        t = item[0]
                        if time.time() < t:
                            # Put the item back into the queue,
                            # it's not time to pop yet
                            self._safe_put(connection_id, item)
                        else:
                            # Try to send the packet
                            packet = item[1]
                            if not packet:
                                # This packet was deleted out from under us
                                continue
                            if packet.status is not Packet.STATUS_ACK:
                                # either need to send initial packet
                                # or resend not-acked packet
                                self._send_packet(packet, connection_id, t)
                    except Empty:
                        pass
                except Exception as e:
                    shared_utils.print_and_log(
                        logging.WARN,
                        'Unexpected error occurred in socket handling thread: '
                        '{}'.format(repr(e)),
                        should_print=True,
                    )
            time.sleep(shared_utils.THREAD_SHORT_SLEEP)

    def open_channel(self, worker_id, assignment_id):
        """
        Opens a channel for a worker on a given assignment, doesn't re-open if the
        channel is already open.
        """
        connection_id = '{}_{}'.format(worker_id, assignment_id)
        if connection_id in self.queues and self.run[connection_id]:
            shared_utils.print_and_log(
                logging.DEBUG, 'Channel ({}) already open'.format(connection_id)
            )
            return
        self.run[connection_id] = True
        self.queues[connection_id] = PriorityQueue()
        self.worker_assign_ids[connection_id] = (worker_id, assignment_id)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.mturk.core.worlds import MTurkOnboardWorld, MTurkTaskWorld
from parlai.core.worlds import validate
from joblib import Parallel, delayed


class MTurkMultiAgentDialogOnboardWorld(MTurkOnboardWorld):
    def parley(self):
        self.mturk_agent.observe({'id': 'System', 'text': 'Welcome onboard!'})
        self.mturk_agent.act()
        self.mturk_agent.observe(
            {
                'id': 'System',
                'text': 'Thank you for your input! Please wait while '
                'we match you with another worker...',
            }
        )
        self.episodeDone = True


class MTurkMultiAgentDialogWorld(MTurkTaskWorld):
    """
    Basic world where each agent gets a turn in a round-robin fashion, receiving as
    input the actions of all other agents since that agent last acted.
    """

    def __init__(self, opt, agents=None, shared=None):
        # Add passed in agents directly.
        self.agents = agents
        self.acts = [None] * len(agents)
        self.episodeDone = False

    def parley(self):
        """
        For each agent, get an observation of the last action each of the other agents
        took.

        Then take an action yourself.
        """
        acts = self.acts
        for index, agent in enumerate(self.agents):
            try:
                acts[index] = agent.act(timeout=None)
            except TypeError:
                acts[index] = agent.act()  # not MTurkAgent
            if acts[index]['episode_done']:
                self.episodeDone = True
            for other_agent in self.agents:
                if other_agent != agent:
                    other_agent.observe(validate(acts[index]))

    def episode_done(self):
        return self.episodeDone

    def shutdown(self):
        """
        Shutdown all mturk agents in parallel, otherwise if one mturk agent is
        disconnected then it could prevent other mturk agents from completing.
        """
        global shutdown_agent

        def shutdown_agent(agent):
            try:
                agent.shutdown(timeout=None)
            except Exception:
                agent.shutdown()  # not MTurkAgent

        Parallel(n_jobs=len(self.agents), backend='threading')(
            delayed(shutdown_agent)(agent) for agent in self.agents
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.worlds import World
from parlai.mturk.core.dev.agents import AssignState


class MTurkDataWorld(World):
    def prep_save_data(self, workers):
        """
        This prepares data to be saved for later review, including chats from individual
        worker perspectives.
        """
        custom_data = self.get_custom_task_data()
        save_data = {'custom_data': custom_data, 'worker_data': {}}

        for agent in workers:
            messages = agent.get_messages()
            # filter out peer feedback
            save_messages = [m for m in messages if m.get('text') != '[PEER_REVIEW]']
            save_data['worker_data'][agent.worker_id] = {
                'worker_id': agent.worker_id,
                'agent_id': agent.id,
                'assignment_id': agent.assignment_id,
                'messages': save_messages,
                'given_feedback': agent.feedback,
                'completed': self.episode_done(),
            }

        # In simple pairing case, attach the feedback right here
        if len(workers) == 2:
            data = save_data['worker_data']
            a_0 = workers[0]
            a_1 = workers[1]
            data[a_0.worker_id]['received_feedback'] = a_1.feedback
            data[a_1.worker_id]['received_feedback'] = a_0.feedback

        return save_data

    def get_custom_task_data(self):
        """
        This function should take the contents of whatever was collected during this
        task that should be saved and return it in some format, preferrably a dict
        containing acts.

        If data needs pickling, put it in a field named 'needs-pickle'
        """
        # return {
        #     'acts': [self.important_turn1, self.important_turn2]
        #     'context': self.some_context_data_of_importance
        #     'needs-pickle': self.json_incompatible_object
        # }
        pass


class MTurkOnboardWorld(MTurkDataWorld):
    """
    Generic world for onboarding a Turker and collecting information from them.
    """

    def __init__(self, opt, mturk_agent):
        """
        Init should set up resources for running the onboarding world.
        """
        self.mturk_agent = mturk_agent
        self.episodeDone = False

    def parley(self):
        """
        A parley should represent one turn of your onboarding task.
        """
        self.episodeDone = True

    def episode_done(self):
        return self.episodeDone

    def review_work(self):
        """
        This call is an opportunity to act on this worker based on their onboarding
        work.

        Generally one could assign a qualification to soft block members who didn't pass
        the onboarding world.
        """
        # if self.episodeDone and work_was_bad:
        #   self.mturk_agent.soft_block_worker()
        #   # Can set a flag for use later in the process
        #   self.mturk_agent.failed_onboarding = True
        pass

    def shutdown(self):
        """
        Clear up resources needed for this world.
        """
        pass


class MTurkTaskWorld(MTurkDataWorld):
    """
    Generic world for MTurk tasks.
    """

    def __init__(self, opt, mturk_agent):
        """
        Init should set up resources for running the task world.
        """
        self.mturk_agent = mturk_agent
        self.episodeDone = False

    def parley(self):
        """
        A parley should represent one turn of your task.
        """
        self.episodeDone = True

    def episode_done(self):
        """
        A ParlAI-MTurk task ends and allows workers to be marked complete when the world
        is finished.
        """
        return self.episodeDone

    def shutdown(self):
        """
        Should be used to free the world's resources and shut down the agents.

        Use the following code if there are multiple MTurk agents:

        global shutdown_agent
        def shutdown_agent(mturk_agent):
            mturk_agent.shutdown()
        Parallel(
            n_jobs=len(self.mturk_agents),
            backend='threading'
        )(delayed(shutdown_agent)(agent) for agent in self.mturk_agents)
        """
        self.mturk_agent.shutdown()

    def review_work(self):
        """
        Programmatically approve/reject the turker's work. Doing this now (if possible)
        means that you don't need to do the work of reviewing later on.

        For example:
        .. code-block:: python
            if self.turker_response == '0':
                self.mturk_agent.reject_work(
                    'You rated our model's response as a 0/10 but we '
                    'know we\'re better than that'
                )
            else:
                if self.turker_response == '10':
                    self.mturk_agent.pay_bonus(1, 'Thanks for a great rating!')
                self.mturk_agent.approve_work()
        """
        # self.mturk_agent.approve_work()
        # self.mturk_agent.reject_work()
        # self.mturk_agent.pay_bonus(1000) # Pay $1000 as bonus
        # self.mturk_agent.block_worker() # Block this worker from future HITs
        pass


class StaticMTurkTaskWorld(MTurkDataWorld):
    """
    World for handling generic tasks that aim to use ParlAI as an MTurk interface, but
    don't need the server to be in the loop.
    """

    def __init__(self, opt, mturk_agent, task_data):
        """
        Init should be provided with the task_data that the worker needs to complete the
        task on the frontend.
        """
        self.mturk_agent = mturk_agent
        self.episodeDone = False
        self.task_data = task_data

    def did_complete(self):
        """
        Determines whether or not this world was completed, or if the agent didn't
        complete the task.
        """
        agent = self.mturk_agent
        return not (agent.hit_is_abandoned or agent.hit_is_returned)

    def episode_done(self):
        """
        A ParlAI-MTurk task ends and allows workers to be marked complete when the world
        is finished.
        """
        return self.episodeDone

    def parley(self):
        """
        A static task parley is simply sending the task data and waiting for the
        response.
        """
        agent = self.mturk_agent
        agent.observe(
            {'id': 'System', 'text': '[TASK_DATA]', 'task_data': self.task_data}
        )
        agent.set_status(AssignState.STATUS_STATIC)
        self.response = agent.get_completed_act()
        self.episodeDone = True

    def prep_save_data(self, workers):
        """
        This prepares data to be saved for later review, including chats from individual
        worker perspectives.
        """
        custom_data = self.get_custom_task_data()
        save_data = {'custom_data': custom_data, 'worker_data': {}}

        agent = self.mturk_agent
        save_data['worker_data'][agent.worker_id] = {
            'worker_id': agent.worker_id,
            'agent_id': agent.id,
            'assignment_id': agent.assignment_id,
            'task_data': self.task_data,
            'response': self.response,
            'completed': self.episode_done(),
        }

        return save_data

    def shutdown(self):
        """
        Shutdown tracking for the agent.
        """
        self.mturk_agent.shutdown()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

import torch
from parlai.agents.hugging_face.dict import Gpt2DictionaryAgent
from parlai.core.torch_generator_agent import TorchGeneratorAgent, TorchGeneratorModel
from parlai.utils.misc import warn_once
from parlai.utils.torch import IdentityLayer, concat_without_padding, padded_tensor


try:
    from transformers import GPT2Model
except ImportError:
    raise ImportError("Please run `pip install transformers`.")


############################################
## Modules
############################################


class GPT2Decoder(torch.nn.Module):
    """
    GPT2 Decoder.

    This decoder is initialized with the pretrained model from Hugging Face.
    """

    def __init__(self, opt, dict):
        super().__init__()
        self.transformer = self._init_from_pretrained(opt)
        # add special tokens
        self.start_idx = dict.start_idx
        self.null_idx = dict.null_idx
        self.add_start_token = False
        if opt["add_special_tokens"]:
            self.transformer.resize_token_embeddings(len(dict.tokenizer))
            self.add_start_token = opt["add_start_token"]
        # use cuda
        self.use_cuda = not opt["no_cuda"] and torch.cuda.is_available()

    def _init_from_pretrained(self, opt):
        # load model
        # check if datapath has the files that hugging face code looks for
        if all(
            os.path.isfile(
                os.path.join(opt["datapath"], "models", "gpt2_hf", file_name)
            )
            for file_name in ["pytorch_model.bin", "config.json"]
        ):
            fle_key = os.path.join(opt["datapath"], "models", "gpt2_hf")
        else:
            model_sz = opt["gpt2_size"]
            if model_sz == "small":
                fle_key = "gpt2"
            elif model_sz == "distilgpt2":
                fle_key = "distilgpt2"
            else:
                fle_key = f"gpt2-{model_sz}"
        return GPT2Model.from_pretrained(fle_key)

    def forward(self, input, encoder_state, incr_state=None):
        attention_mask = None
        if incr_state is None:
            # first step
            if (
                not self.add_start_token
                and input.size(1) == 1
                and int(input[0][0]) == self.start_idx
            ):
                # generating: ignore the start token
                model_input = encoder_state
            else:
                # forced decoding: concatenate the context
                # with the labels
                model_input, _ = concat_without_padding(
                    encoder_state, input, use_cuda=self.use_cuda, null_idx=self.null_idx
                )
                attention_mask = model_input != self.null_idx
        else:
            # generation: get the last token input
            model_input = input[:, -1].unsqueeze(1)

        transformer_outputs = self.transformer(
            model_input, past=incr_state, attention_mask=attention_mask
        )
        hidden_states = transformer_outputs[0]
        new_incr_state = transformer_outputs[1]

        return hidden_states, new_incr_state


class HFGPT2Model(TorchGeneratorModel):
    """
    Hugging Face GPT2 Model.

    GPT2 is a multi-layer decoder-only Transformer. As such, the encoder
    is simply an identity layer. The decoder is initialized with pretrained
    weights from Hugging Face. Read more about this model here
    <https://huggingface.co/transformers/model_doc/gpt2.html>.
    """

    def __init__(self, opt, dict):
        self.null_idx, self.start_idx, self.end_idx = self._get_special_tokens(
            opt, dict
        )
        super().__init__(self.null_idx, self.start_idx, self.end_idx)

        # init the model
        self.encoder = IdentityLayer()
        self.decoder = self._get_decoder(opt, dict)
        self.config = self.decoder.transformer.config
        self.lm_head = torch.nn.Linear(
            self.config.n_embd, self.config.vocab_size, bias=False
        )
        self._tie_weights(self.lm_head, self.decoder.transformer.wte)
        # add start token
        self.add_start_token = opt["add_special_tokens"] and opt["add_start_token"]
        # used to reverse concatenation of context and labels
        self.text_lengths = None

    def _get_decoder(self, opt, dict):
        return GPT2Decoder(opt, dict)

    def _tie_weights(self, output_embeddings, input_embeddings):
        output_embeddings.weight = input_embeddings.weight

    def _get_special_tokens(self, opt, dict):
        return dict.null_idx, dict.start_idx, dict.end_idx

    def reorder_encoder_states(self, encoder_states, indices):
        enc = torch.index_select(encoder_states, 0, indices)
        return enc

    def output(self, tensor):
        """
        Compute output logits.

        Because we concatenate the context with the labels using the
        `concat_without_padding` function, we must truncate the input tensor to return
        only the scores for the label tokens.
        """
        # get only scores for labels
        if self.text_lengths is not None:
            total_length = max(self.text_lengths)
            to_select = tensor.size(1) - total_length
            if not self.add_start_token:
                to_select = to_select + 1
            if to_select > 0:
                # select only label scores
                bsz = tensor.size(0)
                new_tensors = []
                for i in range(bsz):
                    start = self.text_lengths[i]
                    if not self.add_start_token:
                        start = start - 1
                    end = start + to_select
                    new_tensors.append(tensor[i : i + 1, start:end, :])
                tensor = torch.cat(new_tensors, 0)

        return self.lm_head(tensor)

    def reorder_decoder_incremental_state(self, incremental_state, inds):
        new_incr_state = []
        for layer_past in incremental_state:
            new_incr_state.append(torch.index_select(layer_past, 1, inds))

        return tuple(new_incr_state)

    def decode_forced(self, encoder_states, ys):
        """
        Override to get rid of start token input.
        """
        if self.add_start_token:
            return super().decode_forced(encoder_states, ys)
        seqlen = ys.size(1)
        inputs = ys.narrow(1, 0, seqlen - 1)
        latent, _ = self.decoder(inputs, encoder_states)
        logits = self.output(latent)
        _, preds = logits.max(dim=2)
        return logits, preds

    def forward(self, *xs, ys=None, prev_enc=None, maxlen=None, bsz=None):
        model_input, text_lengths = xs
        if ys is not None:
            self.text_lengths = text_lengths
        else:
            self.text_lengths = None

        return super().forward(
            model_input, ys=ys, prev_enc=prev_enc, maxlen=maxlen, bsz=bsz
        )


############################################
## Agent
############################################


class Gpt2Agent(TorchGeneratorAgent):
    """
    Hugging Face GPT2 Agent.

    GPT2 is a multi-layer decoder-only Transformer.
    The decoder is initialized with pretrained weights from Hugging Face.
    Read more about this model here
    <https://huggingface.co/transformers/model_doc/gpt2.html>.

    GPT2 comes in five sizes: distilgpt2, small, medium, large, XL. Use the
    flag `--gpt2-size` to choose the size.

    If you are finetuning the Gpt2 agent as a dialogue agent, be sure
    to run `--add-special-tokens True`. To examine the performance of the
    agent out of the box, run with `--add-special-tokens False`, and make
    sure that the batch size is 1.
    """

    @classmethod
    def add_cmdline_args(cls, argparser):
        agent = argparser.add_argument_group("Gpt2 Args")
        agent.add_argument(
            "--gpt2-size",
            type=str,
            default="small",
            choices=["small", "medium", "large", "xl", "distilgpt2"],
            help="Which size model to initialize.",
        )
        agent.add_argument(
            "--add-special-tokens",
            type="bool",
            default=True,
            help="Add special tokens (like PAD, etc.). If False, "
            "Can only use with batch size 1.",
        )
        agent.add_argument(
            "--add-start-token",
            type="bool",
            default=False,
            help="Add start tokens when finetuning.",
        )
        argparser.set_defaults(
            text_truncate=768,
            label_truncate=256,
            dict_maxexs=0,  # skip building dictionary
        )
        super(Gpt2Agent, cls).add_cmdline_args(argparser)
        warn_once("WARNING: this model is in beta and the API is subject to change.")
        return agent

    def __init__(self, opt, shared=None):
        if not opt["add_special_tokens"] and opt["batchsize"] > 1:
            raise RuntimeError(
                "If using batchsize > 1, --add-special-tokens must be True."
            )
        super().__init__(opt, shared)

    @staticmethod
    def dictionary_class():
        """
        Return the dictionary class that this agent expects to use.

        Can be overriden if a more complex dictionary is required.
        """
        return Gpt2DictionaryAgent

    def build_model(self, states=None):
        """
        Build and return model.
        """
        return HFGPT2Model(self.opt, self.dict)

    def _model_input(self, batch):
        """
        Override to pass in text lengths.
        """
        return (batch.text_vec, batch.text_lengths)

    def _encoder_input(self, batch):
        return (batch.text_vec,)

    def _pad_tensor(self, items):
        """
        Override to always set fp16friendly to False.
        """
        return padded_tensor(
            items, pad_idx=self.NULL_IDX, use_cuda=self.use_cuda, fp16friendly=False
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Count and display statistics of the data.

## Examples

```shell
parlai data_stats -t convai2 -dt train:ordered
```
"""
from parlai.core.params import ParlaiParser
from parlai.agents.repeat_label.repeat_label import RepeatLabelAgent
from parlai.core.worlds import create_task
from parlai.utils.misc import TimeLogger
from parlai.core.dict import DictionaryAgent
from parlai.core.script import ParlaiScript, register_script

import parlai.utils.logging as logging


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, False, 'Compute data statistics')
    # Get command line arguments
    parser.add_argument('-n', '-ne', '--num-examples', type=int, default=-1)
    parser.add_argument('-ltim', '--log-every-n-secs', type=float, default=2)
    parser.add_argument(
        '--agent',
        type=int,
        default=0,
        help='Use teacher (agent 0) or model (agent 1)',
        choices=[0, 1],
    )
    parser.add_argument(
        '--new_line_new_utt',
        type='bool',
        default=False,
        help='New lines treat substrings as separate utterances.',
    )
    parser.add_argument(
        '--ignore_tokens',
        type=str,
        default='',
        help='ignore tokens containings these substrings (comma-separated)',
    )
    parser.set_defaults(datatype='train:ordered')
    DictionaryAgent.add_cmdline_args(parser)
    return parser


def report(world, counts, log_time):
    report = world.report()
    stats = '\n'
    for t in ['input', 'labels', 'both']:
        stats += t + ":\n"
        for s in [
            'utterances_in_',
            'avg_utterance_length_in_',
            'tokens_in_',
            'unique_tokens_in_',
            'unique_utterances_in_',
        ]:
            snice = s.replace('_in_', '').replace('_', ' ')
            stats += "   " + snice + ': ' + str(counts[s + t]) + '\n'
    log = {}
    log['stats'] = stats
    text, log = log_time.log(report['exs'], world.num_examples(), log)
    return text, log


def verify(opt):
    if opt['datatype'] == 'train':
        logging.warn('changing datatype from train to train:ordered')
        opt['datatype'] = 'train:ordered'

    # create repeat label agent and assign it to the specified task
    agent = RepeatLabelAgent(opt)
    world = create_task(opt, agent)
    opt.log()

    log_every_n_secs = opt.get('log_every_n_secs', -1)
    if log_every_n_secs <= 0:
        log_every_n_secs = float('inf')
    log_time = TimeLogger()

    dictionary = DictionaryAgent(opt)
    ignore_tokens = opt.get('ignore_tokens').split(',')

    counts = {}
    for t in {'input', 'labels', 'both'}:
        counts['tokens_in_' + t] = 0
        counts['utterances_in_' + t] = 0
        counts['avg_utterance_length_in_' + t] = 0
        counts['unique_tokens_in_' + t] = 0
        counts['unique_utterances_in_' + t] = 0
        # for counting the stats..
        counts['token_dict_' + t] = {}
        counts['utterance_dict_' + t] = {}

    def tokenize(txt):
        return dictionary.tokenize(txt)

    def keep_token(t):
        for s in ignore_tokens:
            if s != '' and s in t:
                return False
        return True

    # max number of examples to evaluate
    max_cnt = opt['num_examples'] if opt['num_examples'] > 0 else float('inf')
    cnt = 0

    # Show some example dialogs.
    while not world.epoch_done() and cnt < max_cnt:
        cnt += opt.get('batchsize', 1)
        world.parley()
        act = world.get_acts()[opt.get('agent')]
        for itype in {'input', 'labels'}:
            if itype == 'input':
                if opt.get('new_line_new_utt'):
                    txts = act.get('text').split('\n')
                else:
                    txts = [act.get('text')]
            else:
                txts = act.get('labels', act.get('eval_labels', ['']))

            for txt in txts:
                tokens = tokenize(txt)
                retxt = []
                for t in tokens:
                    if keep_token(t):
                        retxt.append(t)
                counts['tokens_in_' + itype] += len(retxt)
                counts['tokens_in_' + 'both'] += len(retxt)
                counts['utterances_in_' + itype] += 1
                counts['utterances_in_' + 'both'] += 1
                counts['avg_utterance_length_in_' + itype] = (
                    counts['tokens_in_' + itype] / counts['utterances_in_' + itype]
                )
                counts['avg_utterance_length_in_' + 'both'] = (
                    counts['tokens_in_' + 'both'] / counts['utterances_in_' + 'both']
                )
                for t in retxt:
                    if t not in counts['token_dict_' + itype]:
                        counts['unique_tokens_in_' + itype] += 1
                        counts['token_dict_' + itype][t] = True
                    if t not in counts['token_dict_' + 'both']:
                        counts['unique_tokens_in_' + 'both'] += 1
                        counts['token_dict_' + 'both'][t] = True
                retxt = ' '.join(retxt)
                if retxt not in counts['utterance_dict_' + itype]:
                    counts['unique_utterances_in_' + itype] += 1
                    counts['utterance_dict_' + itype][retxt] = True
                if retxt not in counts['utterance_dict_' + 'both']:
                    counts['unique_utterances_in_' + 'both'] += 1
                    counts['utterance_dict_' + 'both'][retxt] = True

        if log_time.time() > log_every_n_secs:
            text, log = report(world, counts, log_time)
            logging.info(text)

    try:
        # print dataset size if available
        logging.info(
            f'loaded {world.num_episodes()} episodes with a total '
            f'of {world.num_examples()} examples'
        )
    except Exception:
        pass
    return report(world, counts, log_time)


def obtain_stats(opt, parser):
    report_text, report_log = verify(opt)
    print(report_text.replace('\\n', '\n'))


@register_script('data_stats', hidden=True)
class DataStats(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        return obtain_stats(self.opt, self.parser)


if __name__ == '__main__':
    DataStats.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.params import ParlaiParser
from parlai.mturk.core.mturk_manager import MTurkManager
from worlds import RephrasePersonaWorld, PersonasGenerator
from task_config import task_config

import os


def main():
    """
    This task consists of one agent, model or MTurk worker, talking to an MTurk worker
    to negotiate a deal.
    """
    argparser = ParlaiParser(False, False)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    argparser.add_argument(
        '-min_t', '--min_turns', default=5, type=int, help='minimum number of turns'
    )
    argparser.add_argument(
        '-mt', '--max_turns', default=10, type=int, help='maximal number of chat turns'
    )
    argparser.add_argument(
        '-mx_rsp_time',
        '--max_resp_time',
        default=150,
        type=int,
        help='time limit for entering a dialog message',
    )
    argparser.add_argument(
        '--ag_shutdown_time',
        default=120,
        type=int,
        help='time limit for entering a dialog message',
    )
    argparser.add_argument(
        '--persona-type',
        default='both',
        type=str,
        choices=['both', 'self', 'other'],
        help='Which personas to load from personachat',
    )
    opt = argparser.parse_args()

    directory_path = os.path.dirname(os.path.abspath(__file__))
    opt['task'] = os.path.basename(directory_path)

    opt['extract_personas_path'] = os.path.join(opt['datapath'], opt['task'])
    opt.update(task_config)

    mturk_agent_ids = ['PERSON_1']

    mturk_manager = MTurkManager(opt=opt, mturk_agent_ids=mturk_agent_ids)

    mturk_manager.setup_server(task_directory_path=directory_path)

    personas_generator = PersonasGenerator(opt)
    opt['personas_generator'] = personas_generator

    try:
        mturk_manager.start_new_run()
        mturk_manager.create_hits()

        if not opt['is_sandbox']:
            # ADD BLOCKED WORKERS HERE
            blocked_worker_list = []
            for w in blocked_worker_list:
                mturk_manager.block_worker(
                    w,
                    'We found that you have unexpected behaviors in our '
                    'previous HITs. For more questions please email us.',
                )

        def run_onboard(worker):
            pass

        mturk_manager.set_onboard_function(onboard_function=run_onboard)
        mturk_manager.ready_to_accept_workers()

        def check_worker_eligibility(worker):
            return True

        def assign_worker_roles(workers):
            for index, worker in enumerate(workers):
                worker.id = mturk_agent_ids[index % len(mturk_agent_ids)]

        def run_conversation(mturk_manager, opt, workers):
            worker = workers[0]
            world = RephrasePersonaWorld(opt, worker)
            while not world.episode_done():
                world.parley()
            world.save_data()
            world.shutdown()
            world.review_work()

        mturk_manager.start_task(
            eligibility_function=check_worker_eligibility,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )

    except BaseException:
        raise
    finally:
        mturk_manager.expire_all_unassigned_hits()
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import pickle
import os
from parlai.core.params import ParlaiParser
from parlai.agents.repeat_label.repeat_label import RepeatLabelAgent
from parlai.core.worlds import create_task


def extract_and_save(opt):
    agent = RepeatLabelAgent(opt)
    world = create_task(opt, agent)
    teacher = world.agents[0]

    personas_path = opt.get('personas_path')
    if not os.path.exists(personas_path):
        os.makedirs(personas_path)

    new_episode = True
    personas = []
    while not teacher.epoch_done():
        act = teacher.act()
        if new_episode:
            persona_text = act['text'].split('\n')[:-1]
            if opt.get('persona_type') == 'both':
                persona_1 = [p for p in persona_text if 'your persona:' in p]
                persona_2 = [p for p in persona_text if 'partner\'s persona:' in p]
                persona_1 = [p[p.find(':') + 1 :] for p in persona_1]
                persona_2 = [p[p.find(':') + 1 :] for p in persona_2]
                personas += [persona_1, persona_2]
            else:
                persona = [p for p in persona_text if 'persona:' in p]
                persona = [p[p.find(':') + 1 :] for p in persona]
                personas.append(persona)
            new_episode = act.get('episode_done')
        else:
            new_episode = act.get('episode_done')

    for idx, persona in enumerate(personas):
        with open('{}/{}.pkl'.format(personas_path, idx), 'wb') as f:
            pickle.dump(persona, f)
    print('---Finished extracting and saving personas, to {}'.format(personas_path))


def main(opt):
    print('---Extracting and saving personas---')
    teacher_name = 'personachat:{}'.format(opt.get('persona_type'))
    teacher_name += 'Revised' if opt.get('revised') else 'Original'
    opt['task'] = teacher_name
    assert 'personas_path' in opt, 'Must specify personas path'
    opt['datatype'] = 'train:ordered:stream'
    opt['batchsize'] = 1
    extract_and_save(opt)


if __name__ == '__main__':
    parser = ParlaiParser()
    parser.add_argument(
        '--persona-type',
        default='both',
        type=str,
        choices=['both', 'self', 'other'],
        help='Which personas to load from personachat',
    )
    parser.add_argument(
        '--revised', default=False, type='bool', help='Whether to use revised personas'
    )
    opt = parser.parse_args()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
(A partial) implementation of the DrQa Document Reader from:

Danqi Chen, Adam Fisch, Jason Weston, Antoine Bordes. 2017.
Reading Wikipedia to Answer Open-Domain Questions.
In Association for Computational Linguistics (ACL).

Link: https://arxiv.org/abs/1704.00051

Note:
To use pretrained word embeddings, set the --embedding_file path argument.
GloVe is recommended, see http://nlp.stanford.edu/data/glove.840B.300d.zip.
To automatically download glove, use:
--embedding_file zoo:glove_vectors/glove.840B.300d.txt
"""

try:
    import torch
except ImportError:
    raise ImportError('Need to install pytorch: go to pytorch.org')

import bisect
import numpy as np
import json
import random

from parlai.core.agents import Agent
from parlai.core.dict import DictionaryAgent
from parlai.core.build_data import modelzoo_path
from parlai.utils.io import PathManager
from . import config
from .utils import build_feature_dict, vectorize, batchify, normalize_text
from .model import DocReaderModel


# ------------------------------------------------------------------------------
# Dictionary.
# ------------------------------------------------------------------------------


class SimpleDictionaryAgent(DictionaryAgent):
    """
    Override DictionaryAgent to use spaCy tokenizer.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        group = DictionaryAgent.add_cmdline_args(argparser)
        group.add_argument(
            '--pretrained_words',
            type='bool',
            default=True,
            help='Use only words found in provided embedding_file',
        )
        group.set_defaults(dict_tokenizer='spacy')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Index words in embedding file
        if (
            self.opt['pretrained_words']
            and self.opt.get('embedding_file')
            and not self.opt.get('trained', False)
        ):
            print('[ Indexing words with embeddings... ]')
            self.embedding_words = set()
            self.opt['embedding_file'] = modelzoo_path(
                self.opt.get('datapath'), self.opt['embedding_file']
            )
            with PathManager.open(self.opt['embedding_file']) as f:
                for line in f:
                    w = normalize_text(line.rstrip().split(' ')[0])
                    self.embedding_words.add(w)
            print('[ Num words in set = %d ]' % len(self.embedding_words))
        else:
            self.embedding_words = None

    def add_to_dict(self, tokens):
        """
        Builds dictionary from the list of provided tokens.

        Only adds words contained in self.embedding_words, if not None.
        """
        for token in tokens:
            if self.embedding_words is not None and token not in self.embedding_words:
                continue
            self.freq[token] += 1
            if token not in self.tok2ind:
                index = len(self.tok2ind)
                self.tok2ind[token] = index
                self.ind2tok[index] = token


# ------------------------------------------------------------------------------
# Document Reader.
# ------------------------------------------------------------------------------


class DrqaAgent(Agent):
    @staticmethod
    def add_cmdline_args(argparser):
        config.add_cmdline_args(argparser)
        DrqaAgent.dictionary_class().add_cmdline_args(argparser)

    @staticmethod
    def dictionary_class():
        return SimpleDictionaryAgent

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)

        # All agents keep track of the episode (for multiple questions)
        self.episode_done = True

        self.opt['cuda'] = not self.opt['no_cuda'] and torch.cuda.is_available()

        if shared is not None:
            # model has already been set up
            self.word_dict = shared['word_dict']
            self.model = shared['model']
            self.feature_dict = shared['feature_dict']
        else:
            # set up model
            self.word_dict = DrqaAgent.dictionary_class()(opt)
            if self.opt.get('model_file') and PathManager.exists(opt['model_file']):
                self._init_from_saved(opt['model_file'])
            else:
                if self.opt.get('init_model'):
                    self._init_from_saved(opt['init_model'])
                else:
                    self._init_from_scratch()
            if self.opt['cuda']:
                print('[ Using CUDA (GPU %d) ]' % opt['gpu'])
                torch.cuda.set_device(opt['gpu'])
                self.model.cuda()

        # Set up params/logging/dicts
        self.id = self.__class__.__name__
        config.set_defaults(self.opt)
        self.n_examples = 0

    def _init_from_scratch(self):
        self.feature_dict = build_feature_dict(self.opt)
        self.opt['num_features'] = len(self.feature_dict)
        self.opt['vocab_size'] = len(self.word_dict)

        print('[ Initializing model from scratch ]')
        self.model = DocReaderModel(self.opt, self.word_dict, self.feature_dict)
        self.model.set_embeddings()

    def _init_from_saved(self, fname):
        print('[ Loading model %s ]' % fname)
        saved_params = torch.load(fname, map_location=lambda storage, loc: storage)
        if 'word_dict' in saved_params:
            # for compatibility with old saves
            self.word_dict.copy_dict(saved_params['word_dict'])
        self.feature_dict = saved_params['feature_dict']
        self.state_dict = saved_params['state_dict']
        config.override_args(self.opt, saved_params['config'])
        self.model = DocReaderModel(
            self.opt, self.word_dict, self.feature_dict, self.state_dict
        )

    def share(self):
        shared = super().share()
        shared['word_dict'] = self.word_dict
        shared['model'] = self.model
        shared['feature_dict'] = self.feature_dict
        return shared

    def observe(self, observation):
        # shallow copy observation (deep copy can be expensive)
        observation = observation.copy()
        if not self.episode_done and not observation.get('preprocessed', False):
            dialogue = self.observation['text'].split('\n')[:-1]
            dialogue.extend(observation['text'].split('\n'))
            observation['text'] = '\n'.join(dialogue)
        self.observation = observation
        self.episode_done = observation['episode_done']
        return observation

    def act(self):
        """Update or predict on a single example (batchsize = 1)."""
        reply = {'id': self.getID()}

        ex = self._build_ex(self.observation)
        if ex is None:
            return reply
        batch = batchify(
            [ex], null=self.word_dict[self.word_dict.null_token], cuda=self.opt['cuda']
        )

        # Either train or predict
        if 'labels' in self.observation:
            self.n_examples += 1
            self.model.update(batch)
        else:
            prediction, score = self.model.predict(batch)
            reply['text'] = prediction[0]
            reply['text_candidates'] = [prediction[0]]
            reply['candidate_scores'] = [score[0]]

        reply['metrics'] = {'train_loss': self.model.train_loss.avg}
        return reply

    def batch_act(self, observations):
        """
        Update or predict on a batch of examples.

        More efficient than act().
        """
        batchsize = len(observations)
        batch_reply = [{'id': self.getID()} for _ in range(batchsize)]

        # Some examples will be None (no answer found). Filter them.
        examples = [self._build_ex(obs) for obs in observations]
        valid_inds = [i for i in range(batchsize) if examples[i] is not None]
        examples = [ex for ex in examples if ex is not None]

        # If all examples are invalid, return an empty batch.
        if len(examples) == 0:
            return batch_reply

        # Else, use what we have (hopefully everything).
        batch = batchify(
            examples,
            null=self.word_dict[self.word_dict.null_token],
            cuda=self.opt['cuda'],
        )

        # Either train or predict
        if 'labels' in observations[0]:
            try:
                self.n_examples += len(examples)
                self.model.update(batch)
            except RuntimeError as e:
                # catch out of memory exceptions during fwd/bck (skip batch)
                if 'out of memory' in str(e):
                    print(
                        '| WARNING: ran out of memory, skipping batch. '
                        'if this happens frequently, decrease batchsize or '
                        'truncate the inputs to the model.'
                    )
                    batch_reply[0]['metrics'] = {'skipped_batches': 1}
                    return batch_reply
                else:
                    raise e

        else:
            predictions, scores = self.model.predict(batch)
            for i in range(len(predictions)):
                batch_reply[valid_inds[i]]['text'] = predictions[i]
                batch_reply[valid_inds[i]]['text_candidates'] = [predictions[i]]
                batch_reply[valid_inds[i]]['candidate_scores'] = [scores[i]]

        batch_reply[0]['metrics'] = {
            'train_loss': self.model.train_loss.avg * batchsize
        }
        return batch_reply

    def save(self, fname=None):
        """
        Save the parameters of the agent to a file.
        """
        fname = self.opt.get('model_file', None) if fname is None else fname
        if fname:
            print("[ saving model: " + fname + " ]")
            self.opt['trained'] = True
            self.model.save(fname)
            # save opt file
            with PathManager.open(fname + '.opt', 'w') as handle:
                json.dump(self.opt, handle)

    # --------------------------------------------------------------------------
    # Helper functions.
    # --------------------------------------------------------------------------

    def _build_ex(self, ex):
        """
        Find the token span of the answer in the context for this example.

        If a token span cannot be found, return None. Otherwise, torchify.
        """
        # Check if empty input (end of epoch)
        if 'text' not in ex:
            return

        # Split out document + question
        inputs = {}
        fields = ex['text'].strip().split('\n')

        # Data is expected to be text + '\n' + question
        if len(fields) < 2:
            raise RuntimeError('Invalid input. Is task a QA task?')

        paragraphs, question = fields[:-1], fields[-1]

        if len(fields) > 2 and self.opt.get('subsample_docs', 0) > 0 and 'labels' in ex:
            paragraphs = self._subsample_doc(
                paragraphs, ex['labels'], self.opt.get('subsample_docs', 0)
            )

        document = ' '.join(paragraphs)
        inputs['document'], doc_spans = self.word_dict.span_tokenize(document)
        inputs['question'] = self.word_dict.tokenize(question)
        inputs['target'] = None

        # Find targets (if labels provided).
        # Return if we were unable to find an answer.
        if 'labels' in ex:
            if 'answer_starts' in ex:
                # randomly sort labels and keep the first match
                labels_with_inds = list(zip(ex['labels'], ex['answer_starts']))
                random.shuffle(labels_with_inds)
                for ans, ch_idx in labels_with_inds:
                    # try to find an answer_start matching a tokenized answer
                    start_idx = bisect.bisect_left(
                        list(x[0] for x in doc_spans), ch_idx
                    )
                    end_idx = start_idx + len(self.word_dict.tokenize(ans)) - 1
                    if end_idx < len(doc_spans):
                        inputs['target'] = (start_idx, end_idx)
                        break
            else:
                inputs['target'] = self._find_target(inputs['document'], ex['labels'])
            if inputs['target'] is None:
                return

        # Vectorize.
        inputs = vectorize(self.opt, inputs, self.word_dict, self.feature_dict)

        # Return inputs with original text + spans (keep for prediction)
        return inputs + (document, doc_spans)

    def _find_target(self, document, labels):
        """
        Find the start/end token span for all labels in document.

        Return a random one for training.
        """

        def _positions(d, l):
            for i in range(len(d)):
                for j in range(i, min(len(d) - 1, i + len(l))):
                    if l == d[i : j + 1]:
                        yield (i, j)

        targets = []
        for label in labels:
            targets.extend(_positions(document, self.word_dict.tokenize(label)))
        if len(targets) == 0:
            return
        return targets[np.random.choice(len(targets))]

    def _subsample_doc(self, paras, labels, subsample):
        """
        Subsample paragraphs from the document (mostly for training speed).
        """
        # first find a valid paragraph (with a label)
        pi = -1
        for ind, p in enumerate(paras):
            for l in labels:
                if p.find(l):
                    pi = ind
                    break
        if pi == -1:
            # failed
            return paras[0:1]
        new_paras = []
        if pi > 0:
            for _i in range(min(subsample, pi - 1)):
                ind = random.randint(0, pi - 1)
                new_paras.append(paras[ind])
        new_paras.append(paras[pi])
        if pi < len(paras) - 1:
            for _i in range(min(subsample, len(paras) - 1 - pi)):
                ind = random.randint(pi + 1, len(paras) - 1)
                new_paras.append(paras[ind])
        return new_paras



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import sys

import parlai.mturk.core.mturk_utils as mturk_utils


def main():
    """
    This script should be used after some error occurs that leaves HITs live while the
    ParlAI MTurk server down.

    This will search through live HITs and list them by task ID, letting you close down
    HITs that do not link to any server and are thus irrecoverable.
    """
    parser = argparse.ArgumentParser(description='Delete HITs by expiring')
    parser.add_argument(
        '--sandbox',
        dest='sandbox',
        default=False,
        action='store_true',
        help='Delete HITs from sandbox',
    )
    parser.add_argument(
        '--ignore-assigned',
        dest='ignore',
        default=False,
        action='store_true',
        help='Ignore HITs that may already be completed or assigned',
    )
    parser.add_argument(
        '--approve',
        dest='approve',
        default=False,
        action='store_true',
        help='Approve HITs that have been completed on deletion',
    )
    parser.add_argument(
        '--verbose',
        dest='verbose',
        default=False,
        action='store_true',
        help='List actions to individual HITs',
    )
    opt = parser.parse_args()
    sandbox = opt.sandbox
    ignore = opt.ignore
    approve = opt.approve
    verbose = opt.verbose
    task_group_ids = []
    group_to_hit = {}
    processed = 0
    found = 0
    spinner_vals = ['-', '\\', '|', '/']
    if sandbox:
        print(
            'About to query the SANDBOX server, these HITs will be active HITs'
            ' from within the MTurk requester sandbox'
        )
    else:
        print(
            'About to query the LIVE server, these HITs will be active HITs '
            'potentially being worked on by real Turkers right now'
        )

    print(
        'Getting HITs from amazon MTurk server, please wait...\n'
        'or use CTRL-C to skip to expiring some of what is found.\n'
    )
    mturk_utils.setup_aws_credentials()
    client = mturk_utils.get_mturk_client(sandbox)
    response = client.list_hits(MaxResults=100)
    try:
        while True:
            processed += response['NumResults']
            for hit in response['HITs']:
                if ignore:
                    if hit['NumberOfAssignmentsAvailable'] == 0:
                        # Ignore hits with no assignable assignments
                        continue
                    if (
                        hit['HITStatus'] != 'Assignable'
                        and hit['HITStatus'] != 'Unassignable'
                    ):
                        # Ignore completed hits
                        continue
                question = hit['Question']
                try:
                    if 'ExternalURL' in question:
                        url = question.split('ExternalURL')[1]
                        group_id = url.split('task_group_id=')[1]
                        group_id = group_id.split('&')[0]
                        group_id = group_id.split('<')[0]
                        if group_id not in task_group_ids:
                            sys.stdout.write(
                                '\rFound group {}                         '
                                '                                         '
                                '\n'.format(group_id)
                            )
                            group_to_hit[group_id] = {}
                            task_group_ids.append(group_id)
                        group_to_hit[group_id][hit['HITId']] = hit['HITStatus']
                        found += 1
                except IndexError:
                    pass  # This wasn't the right HIT

            sys.stdout.write(
                '\r{} HITs processed, {} active hits'
                ' found amongst {} tasks. {}        '.format(
                    processed,
                    found,
                    len(task_group_ids),
                    spinner_vals[((int)(processed / 100)) % 4],
                )
            )
            if 'NextToken' not in response:
                break
            response = client.list_hits(NextToken=response['NextToken'], MaxResults=100)

    except BaseException as e:
        print(e)
        pass

    if not approve:
        print('\n\nTask group id - Active HITs')
        for group_id in task_group_ids:
            print('{} - {}'.format(group_id, len(group_to_hit[group_id])))
    else:
        print('\n\nTask group id - Active HITs - Reviewable')
        for group_id in task_group_ids:
            print(
                '{} - {} - {}'.format(
                    group_id,
                    len(group_to_hit[group_id]),
                    len(
                        [
                            s
                            for s in group_to_hit[group_id].values()
                            if s == 'Reviewable'
                        ]
                    ),
                )
            )

    print(
        'To clear a task, please enter the task group id of the task that you '
        'want to expire the HITs for. To exit, enter nothing'
    )

    while True:
        task_group_id = input("Enter task group id: ")
        if len(task_group_id) == 0:
            break
        elif task_group_id not in task_group_ids:
            print('Sorry, the id you entered was not found, try again')
        else:
            num_hits = input(
                'Confirm by entering the number of HITs that will be deleted: '
            )
            if '{}'.format(len(group_to_hit[task_group_id])) == num_hits:
                hits_expired = 0
                for hit_id, status in group_to_hit[task_group_id].items():
                    if approve:
                        response = client.list_assignments_for_hit(HITId=hit_id)
                        if response['NumResults'] == 0:
                            if verbose:
                                print(
                                    'No results for hit {} with status {}\n'
                                    ''.format(hit_id, status)
                                )
                        else:
                            assignment = response['Assignments'][0]
                            assignment_id = assignment['AssignmentId']
                            client.approve_assignment(AssignmentId=assignment_id)
                            if verbose:
                                print('Approved assignment {}'.format(assignment_id))
                    mturk_utils.expire_hit(sandbox, hit_id)
                    if verbose:
                        print('Expired hit {}'.format(hit_id))
                    hits_expired += 1
                    sys.stdout.write('\rExpired hits {}'.format(hits_expired))
                print(
                    '\nAll hits for group {} have been expired.'.format(task_group_id)
                )
            else:
                print(
                    'You entered {} but there are {} HITs to expire, please '
                    "try again to confirm you're ending the right task".format(
                        num_hits, len(group_to_hit[task_group_id])
                    )
                )


if __name__ == '__main__':
    main()
